"""
Container classes for variables
"""

import numpy as np
import os
import pandas as pd
import tables
import abc
import enum
import array
import numbers
from ..utils import files as f
from ..utils import Logger
from copy import deepcopy as copy
DEFAULT_BINS = 70

REGISTERED_FILEEXTENSIONS = [".h5"]

try:
    import uproot as ur
    import uproot_methods.classes.TVector3 as TVector3
    import uproot_methods.classes.TLorentzVector as TLorentzVector
    import uproot_methods.classes.TH1
    REGISTERED_FILEEXTENSIONS.append(".root")

except ImportError:
    Logger.warning("No uproot found, root support is limited!")


################################################################
# define a non-member function so that it can be used in a
# multiprocessing approach
def extract_from_root(filename, definitions,
                      nevents=None,
                      dtype=np.float64,
                      transform = None,
                      reduce_dimension=None):
    """
    Use the uproot system to get information from rootfiles. Supports a basic tree of
    primitive datatype like structure.

    Args:
        filename (str): datafile
        definitiions (list): tree and branch adresses

    Keyword Args:
        nevents (int): number of events to read out
        reduce_dimension (int): If data is vector-type, reduce it by taking the n-th element
        dtype (np.dtyoe):  A numpy datatype, default double (np.float64) - use smaller dtypes to 
                           save memory
        transform (func): A function which directy transforms the readout data
    """

    can_be_concatted = False
    rootfile = ur.open(filename)
    success = False
    i=0
    branch = None
    # it will most likely work only with TTrees
    while not success:
        try:
            tree = rootfile.get(definitions[i][0])
            branch = tree
            for address in definitions[i][1:]:
                Logger.debug("Searching for address {}".format(address))
                branch = branch.get(address)

            #tree = file[definitions[i][0]]
            #branch = rootfile[definitions[i][0]].get(definitions[i][1])
            success = True
        except KeyError as e:
            Logger.warning(f"Can not find address {definitions[i]}")
            i+=1
        except IndexError:
            Logger.critical(f"None of the provided keys could be found {definitions}")
            break
    Logger.debug(f"Found valid definitions {definitions[i]}")

    ##FiXME make logger.critical end program!
    if nevents is not None:
        data = branch.array(entrystop=nevents)
    else:
        data = branch.array()

    # check for dimensionality
    multidim = False
    try:
        len(data[0])
        multidim = True
    except TypeError:
        Logger.debug(f"Assuming scalar data {definitions[i]}")

    if multidim:
        Logger.debug("Inspecting data...")
        tmp_lengths = set([len(k) for k in data])
        Logger.debug("Found {}".format(tmp_lengths))
        firstlen = list(tmp_lengths)[0]
        if (len(tmp_lengths) == 1) and (firstlen == 1):
            multidim = False
            Logger.debug("Found data containing iterables of size 1... flattening!")
            del tmp_lengths
            if dtype != np.float64:
                tmpdata = array.array("f",[])
            else:
                tmpdata = array.array("d",[])
            if isinstance(data[0][0], numbers.Number):
                [tmpdata.append(dtype(k)) for k in data]
                #tmpdata = np.asarray([k[0] for k in data])
                #multidim = True
                data = tmpdata
                del tmpdata 
            else:
                Logger.info("Is multidim data")
                multidim = True
        else:
            del tmp_lengths
            multidim = True
            Logger.debug("Assuming array data {}".format(definitions[i]))

    if reduce_dimension is not None:
        if not multidim:
            raise ValueError("Can not reduce scalar data!")
        if isinstance(reduce_dimension, int):
            data = np.array([k[reduce_dimension] for k in data], dtype=dtype)
            multidim = False
        else:
            data = [[k[reduce_dimension[1]] for k in j] for j in data]
    if multidim:
        Logger.debug("Grabbing multidimensional data from root-tree for {}".format(definitions[i]))
        del data
        if nevents is None:
            data = branch.array() #this will be a jagged array now!
        else:
            data = branch.array(entrystop=nevents)
        del branch
        if (len(data[0])):
            if isinstance(data[0][0], TVector3.TVector3):
                Logger.info("Found TVector3 data, treating appropriatly")
                data =  pd.Series([np.array([i.x,i.y,i.z], dtype=dtype) for i in data])
            if isinstance(data[0][0], TLorentzVector.TLorentzVector):
                Logger.info("Found TLorentzVector data, treating appropriatly")
                data =  pd.Series([np.array([i.x,i.y,i.z,i.t], dtype=dtype) for i in data])
        else: # probably number then    
            data = pd.Series([np.asarray(i,dtype=dtype) for i in data])
    
        # the below might not be picklable (multiprocessing!)
        #tmpdata = [i for i in data]
        # FIXME: dataframe/series
        # try to convert this to a pandas dataframe
        #data = pd.DataFrame(tmpdata)
        can_be_concatted = True
        data.multidim = True
    else:
        Logger.debug("Grabbing scalar data from root-tree for {}".format(definitions[i]))
        # convert in cases of TVector3/TLorentzVector

        if isinstance(data[0], TVector3.TVector3):
            Logger.debug("Found TVector3")
            data =  pd.Series([np.array([i.x,i.y,i.z], dtype=dtype) for i in data])
        elif isinstance(data[0], TLorentzVector.TLorentzVector):
            Logger.debug("Found TLorentzVector")
            data =  pd.Series([np.array([i.x,i.y,i.z, i.t], dtype=dtype) for i in data])
        else:
            try:
                #FIXME: why is that asarray needed?
                #data = pd.Series(np.asarray(data,dtype=dtype))
                data = pd.Series(data,dtype=dtype)
            except TypeError: # data consist of some object
                data = pd.Series(data) 
        Logger.debug("Got {} elements for {}".format(len(data), definitions[i]))
        can_be_concatted = True
    if transform is not None:
        data = transform(data)
    return data, can_be_concatted

################################################################
# define a non-member function so that it can be used in a
# multiprocessing approach
def harvest(filenames, definitions, **kwargs):
    """
    Read variables from files into memory. Will be used by HErmes.selection.variables.Variable.harvest
    This will be run multi-threaded. Keep that in mind, arguments have to be picklable,
    also everything thing which is read out must be picklable. Lambda functions are NOT picklable

    Args:
        filenames (list): the files to extract the variables from.
                          currently supported: hdf
        definitions (list): where to find the data in the files. They usually
                            have some tree-like structure, so this a list
                            of leaf-value pairs. If there is more than one
                            all of them will be tried. (As it might be that
                            in some files a different naming scheme was used)
                            Example: [("hello_reoncstruction", "x"), ("hello_reoncstruction", "y")] ]
    Keyword Args:
        transformation (func): After the data is read out from the files,
                               transformation will be applied, e.g. the log
                               to the energy.
        fill_empty (bool): Fill empty fields with zeros
        nevents (int): ROOT only - read out only nevents from the files
        reduce_dimension (str): ROOT only - multidimensional data can be reduced by only
                                            using the index given by reduce_dimension.
                                            E.g. in case of a TVector3, and we want to have onlz
                                            x, that would be 0, y -> 1 and z -> 2.
        dtype (np.dtype) : datatype to cast to (default np.float64, but can be used
                           to reduce memory footprint.

    Returns:
        pd.Series or pd.DataFrame
    """

    nevents          = kwargs["nevents"] if "nevents" in kwargs else None
    fill_empty       = kwargs["fill_empty"] if "fill_empty" in kwargs else False
    reduce_dimension = kwargs["reduce_dimension"] if "reduce_dimension" in kwargs else None
    transform        = kwargs["transformation"] if "transformation" in kwargs else None
    dtype            = kwargs["dtype"] if "dtype" in kwargs else np.float64

    concattable = True
    data = pd.Series(dtype=dtype)
    #multidim_data = pd.DataFrame()
    for filename in filenames:
        filetype = f.strip_all_endings(filename)[1]

        assert filetype in REGISTERED_FILEEXTENSIONS, "Filetype {} not known!".format(filetype)
        assert os.path.exists(filename), "File {} does not exist!".format(filetype)
        if (filetype == ".h5") and (transform is not None):
            Logger.critical("Can not apply direct transformation for h5 files (yet). This is only important for root files and varaibles which are used as VariableRole.PARAMETER")
        Logger.debug("Attempting to harvest {1} file {0}".format(filename,filetype))
        
        if filetype == ".h5" and not isinstance(filename, tables.table.Table):
            # store = pd.HDFStore(filename)
            hdftable = tables.open_file(filename)

        else:
            hdftable = filename

        tmpdata = pd.Series(dtype=dtype)
        for definition in definitions:

            definition = list(definition) 
            if filetype == ".h5":
                if not definition[0].startswith("/"):
                    definition[0] = "/" + definition[0]
                try:
                    # data = store.select_column(*definition)
                    if not definition[1]:
                        tmpdata = hdftable.get_node(definition[0])
                    else:
                        tmpdata = hdftable.get_node(definition[0]).col(definition[1])
                    if tmpdata.ndim == 2:
                        if data.empty:
                            data = pd.DataFrame()
                        tmpdata = pd.DataFrame(tmpdata, dtype=dtype)
                    else:
                        tmpdata = pd.Series(tmpdata, dtype=dtype)

                    Logger.debug("Found {} entries in table for {}{}".format(len(tmpdata),definition[0],definition[1]))
                    break
                except tables.NoSuchNodeError:
                    Logger.debug("Can not find definition {0} in {1}! ".format(definition, filename))
                    continue

            elif filetype == ".root":
                tmpdata, concattable = extract_from_root(filename, definitions,
                                                         nevents=nevents,
                                                         dtype=dtype,
                                                         transform=transform,
                                                         reduce_dimension=reduce_dimension)
        if filetype == ".h5":
            hdftable.close()

        #tmpdata = harvest_single_file(filename, filetype,definitions)
        # self.data = self.data.append(data.map(self.transform))
        # concat should be much faster
        if not True in [isinstance(tmpdata, k) for k in [pd.Series, pd.DataFrame] ]:
            concattable = False

        if not concattable:
            Logger.warning(f"Data {definitions} can not be concatted, keep that in mind!")
            try:
                tmpdata = pd.Series(tmpdata)
                #return tmpdata
            except:
                tmpdata = [k for k in tmpdata]
                tmpdata = pd.Series(tmpdata)
                #return tmpdata

        data = pd.concat([data, tmpdata])

        del tmpdata
    return data

################################################################

def freedman_diaconis_bins(data,leftedge,\
                           rightedge,minbins=20,\
                           maxbins=70,fallbackbins=DEFAULT_BINS):
    """
    Get a number of bins for a histogram
    following Freedman/Diaconis

    Args:
        leftedge (float): left bin edge
        rightedge (float): right bin edge
        minbins (int): the minimum number of bins
        maxbins (int): the maximum number of bins
        fallbackbins (int): a number of bins which is returned
                            if calculation failse

    Returns:
        nbins (int): number of bins, minbins < bins < maxbins
    """

    try:
        finite_data = np.isfinite(data)
        q3          = np.percentile(data[finite_data],75)
        q1          = np.percentile(data[finite_data],25)
        n_data      = len(data)
        if q3 == q1:
            Logger.warning("Can not calculate bins, falling back... to min max approach")
            q3 = max(finite_data)
            q1 = min(finite_data)

        h           = (2*(q3-q1))/(n_data**1./3)
        bins = (rightedge - leftedge)/h


        if not np.isfinite(bins):
            Logger.info(f"Got some nan somewhere: q1 : {q1}, q3 : {q3}, n_data : {n_data}, h : {h}")
            Logger.warning("Calculate Freedman-Draconis bins failed, calculated nan bins, returning fallback")
            bins = fallbackbins

        if bins < minbins:
            bins = minbins
        if bins > maxbins:
            bins = maxbins
    except Exception as e:
        Logger.warning(f"Calculate Freedman-Draconis bins failed {e}")
        bins = fallbackbins
    return bins

#############################################################

class VariableRole(enum.Enum):
    """
    Define roles for variables. Some variables used in a special context (like weights)
    are easily recognizable by this flag.
    """
    UNKNOWN         = 0
    SCALAR          = 10
    ARRAY           = 20
    GENERATORWEIGHT = 30
    RUNID           = 40
    EVENTID         = 50
    STARTIME        = 60
    ENDTIME         = 70
    FLUXWEIGHT      = 80
    PARAMETER       = 90 # a single parameter, no array whatsoever


##############################################################

class AbstractBaseVariable(metaclass=abc.ABCMeta):
    """
    Read out tagged numerical data from files
    """    
    _harvested = False
    _bins = None
    ROLES = VariableRole

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        return """<Variable: {}>""".format(self.name)

    def __eq__(self,other):
        return self.name == other.name

    def __lt__(self, other):
        return sorted([self.name,other.name])[0] == self.name

    def __gt__(self, other):
        return sorted([self.name,other.name])[1] == self.name

    def __le__(self, other):
        return self < other or self == other

    def __ge__(self, other):
        return self > other or self == other

    def declare_harvested(self):
        self._harvested = True

    @property
    def harvested(self):
        return self._harvested

    @property
    def bins(self):
        if self._bins is None:
            return self.calculate_fd_bins()
        else:
            return self._bins

    @bins.setter
    def bins(self, value):
        self._bins = value

    def calculate_fd_bins(self, cutmask=None):
        """
        Calculate a reasonable binning

        Keyword Args:
            cutmask (numpy.ndarray) : a boolean mask to cut on, in case 
                                      cuts have been applied to the 
                                      category this data is part of

        Returns:
            numpy.ndarray: Freedman Diaconis bins
        """
        tmpdata = self.data
        if cutmask is not None:
            if len(cutmask) > 0:
                tmpdata = tmpdata[cutmask]

        nbins = freedman_diaconis_bins(tmpdata, min(tmpdata), max(tmpdata))
        bins = np.linspace(min(tmpdata),max(tmpdata), nbins)
        return bins

    def harvest(self, *files):
        """
        Hook to the harvest method. Don't use in case of multiprocessing!
        Args:
            *files: walk through these files and readout
        """
        if self.role == VariableRole.PARAMETER:
            self._data = harvest(files, self.definitions, transformation= self.transform)
            self._data = self._data[0]
        else:
            self._data = harvest(files, self.definitions)
        self.declare_harvested()

    @abc.abstractmethod
    def rewire_variables(self, vardict):
        return

    @property
    def ndim(self):
        if hasattr(self._data, "multidim"):
            if self._data.multidim == True:
                return 2
        return self._data.ndim

    @property
    def data(self):
        if isinstance(self._data, pd.DataFrame):
            #return self._data.as_matrix()
            #FIXME: as_matrix is depracted in favor of values
            return self._data.values
        if not hasattr(self._data, "shape"):
            Logger.warning("Something's wrong, this should be array data!")
            Logger.warning(f"Seeeing {type(self._data)} data")
            Logger.warning("Attempting to fix!")
            self._data = np.asarray(self._data)

        return self._data

############################################

class Variable(AbstractBaseVariable):
    """
    A hook to a single variable read out from a file
    """

    def __init__(self, name, definitions=None,\
                 bins=None, label="", transform=None,
                 role=VariableRole.SCALAR,
                 nevents=None,
                 reduce_dimension=None):
        """
        Args:
            name                                     (str) : An unique identifier

        Keyword Args:
            definitions                             (list) : table and/or column names in underlying data
            bins                           (numpy.ndarray) : used for histograms
            label                                    (str) : used for plotting and as a label in tables
            transform                               (func) : apply to each member of the underlying data at readout
            role (HErmes.selection.variables.VariableRole) : The role the variable is playing. 
                                                             In most cases the default is the best choice
            nevents                                  (int) : number of events to read in (ROOT only right now!)
            reduce_dimension                         (int) : in case of multidimensionality,
                                                             take only the the given index of the array (ROOT only right now)
        """
        AbstractBaseVariable.__init__(self)

        if definitions is not None:
            #assert not (False in [len(x) <= 2 for x in definitions]), "Can not understand variable definitions {}!".format(definitions)
            self.defsize = len(definitions[0])
            #FIXME : not sure how important this is right now
            #assert not (False in [len(x) == self.defsize for x in definitions]), "All definitions must have the same length!"
        else:
            self.defsize = 0

        self.name        = name
        self.role        = role
        self.bins        = bins # when histogrammed
        self.label       = label
        self.transform   = transform
        self.definitions = definitions
        self._data       = pd.Series(dtype=np.float64)
        self.nevents     = nevents
        self.reduce_dimension = reduce_dimension
        self._role = role

        #if self.defsize  == 1:
        #    self.data    = pd.DataFrame()
        #if self.defsize  == 2:
        #    self.data    = pd.Series()

    def rewire_variables(self, vardict):
        """
        Make sure all the variables are connected properly. This is
        only needed for combined/compound variables

        Returns:
            None
        """
        pass

##########################################################


class CompoundVariable(AbstractBaseVariable):
    """
    Calculate a variable from other variables. This kind of variable will not read any file.
    """

    def __init__(self, name, variables=None, label="",\
                 bins=None, operation=lambda x,y : x + y,
                 role=VariableRole.SCALAR,
                 dtype=np.float64):
        """
        A compound variable is a variable which is created from two or more other variables. This variable does not have
        a direct representation in a file, but gets calculated on the fly instead, e.g. a residual of two other variables
        The 'operation' denoted function here defines what operator should be applied to the variables to create the new
        coumpound variable

        Args:
            name                                     (str) : An unique identifier for the new variable.

        Keyword Args:
            variables                               (list) : A list of variables used to calculate the new variable.
            label                                    (str) : A label for plotting.
            bins                              (np.ndarray) : binning for distributions.
            operation                                (fnc) : The operation which will be applied to variables.
            role (HErmes.selection.variables.VariableRole) : The role the variable is playing.
                                                             In most cases the default is the best choice. Assigning roles
                                                             to variables allows for special magic, e.g. in the case
                                                             of weighting
        """
        AbstractBaseVariable.__init__(self)
        self.name = name
        self.role = role
        self.label = label
        self.bins = bins
        if variables is None:
            variables = []
        self.variables = variables
        self.operation = operation
        self._data = pd.Series(dtype=np.float64) #dtype to suppress warning
        self.definitions = ((self.__repr__()),)

    def rewire_variables(self, vardict):
        """
        Use to avoid the necessity to read out variables twice
        as the variables are copied over by the categories, 
        the refernce is lost. Can be rewired though
        """
        newvars = []
        for var in self.variables:
            newvars.append(vardict[var.name])
        self.variables = newvars

    def __repr__(self):
        return """<CompoundVariable {} created from: {}>""".format(self.name,"".join([x.name for x in self.variables ]))

    def harvest(self, *filenames):
        #FIXME: filenames is not used, just
        #there for compatibility

        if self.harvested:
            return
        harvested = [var for var in self.variables if var.harvested]
        if not len(harvested) == len(self.variables):
            Logger.error("Variables have to be harvested for compound variable {0} first!".format(self.variables))
            Logger.error("Only {} is harvested".format(harvested))
            return
        self._data = self.operation(*[var.data for var in self.variables])
        self.declare_harvested()

##########################################################


class VariableList(AbstractBaseVariable):
    """
    A list of variable. Can not be read out from files.
    """

    def __init__(self, name, variables=None, label="", bins=None, role=VariableRole.SCALAR):
        """
        Args:
            name (str): An unique identifier for the new category.

        Keyword Args:
            variables (list): A list of variables used to calculate the new variable.
            label (str): A label for plotting.
            bins (np.ndarray): binning for distributions.
            role (HErmes.selection.variables.VariableRole): The role the variable is playing. In most cases the default is the best choice
        """

        AbstractBaseVariable.__init__(self)
        self.name = name
        self.label = label
        self.bins = bins
        if variables is None:
            variables = []
        self.variables = variables

    def harvest(self, *filenames):
        #FIXME: filenames is not used, just
        #there for compatibility

        # do not calculate weights yet!


        if self.harvested:
            return
        harvested = [var for var in self.variables if var.harvested]
        if not len(harvested) == len(self.variables):
            Logger.error("Variables have to be harvested for compound variable {} first!".format(self.name))
            return
        self.declare_harvested()

    def rewire_variables(self, vardict):
        """
        Use to avoid the necessity to read out variables twice
        as the variables are copied over by the categories,
        the refernce is lost. Can be rewired though
        """
        newvars = []
        for var in self.variables:
            newvars.append(vardict[var.name])
        self.variables = newvars

    @property
    def data(self):
        return [x.data for x in self.variables]






"""
All magic keywords shall summon here
"""

MC_P_EN   = "mc_p_en"
MC_P_TY   = "mc_p_ty"
MC_P_ZE   = "mc_p_ze"
MC_P_WE   = "mc_p_we"
MC_P_GW   = "mc_p_gw"
MC_P_TS   = "mc_p_ts"
RUN_START = "run_start_mjd"
RUN_STOP  = "run_stop_mjd"
RUN       = "run"
EVENT     = "event"
DATASETS  = "dataset"



"""
Miscellaneous tools
"""


__all__ = ["files"]


from hepbasestack import logger
from hepbasestack.itools import slicer, flatten, multiplex
from hepbasestack import timeit, isnotebook

from . import files

Logger = logger.Logger




"""
Flux models for atmospheric neutrino and muon fluxes
as well as power law fluxes
"""
from __future__ import absolute_import

import numpy as np
from functools import reduce
from . import conversions as conv

icecube_found = False
try:
    from icecube import icetray, dataclasses, NewNuFlux
    from icecube.weighting.weighting import from_simprod
    import icecube.weighting.fluxes as ICMuFluxes
    icecube_found = True
except ImportError:
    print("WARNING: module icecube not found!")

    # hobo mufluxes
    class ICMuFluxes(object):
        GaisserH3a = None
        GaisserH4a = None
        Hoerandel  = None
        Hoerandel5 = None

    def AtmosphericNuFlux(*args, **kwargs):
        hobo =  lambda x: x
        return hobo

AREA_SUM = 18946832.9035663

###############################################


def PowerLawFlux(fluxconst=1e-8,gamma=2):
    """
    A simple powerlaw flux

    Args:
        fluxconst (float): normalization
        gamma (float): spectral index

    Returns (func): the flux function

    """

    if gamma > 0:
        gamma *= -1

    def flux(mc_p_energy,mc_p_type,mc_p_zenith, fluxconst=fluxconst, gamma=gamma):
        # weighting API requires second and third argument even if we
        # don't need it here
        return fluxconst * np.power(mc_p_energy, gamma)

    return flux

###############################################

if icecube_found:

    def AtmosphericNuFlux(modelname='honda2006',knee=False, fluxconst=1.):
        """
        Create an Atmospheric neutrino flux

        Args:
            modelname (str): a atmospheric model
            knee (bool): ad a knee model
            fluxconst (float): scale flux by this const

        Returns (func): the requested flux. Takes energy, type and maybe cos(zenith) as parameters

        """
        nuflux = NewNuFlux.makeFlux(modelname)
        if knee:
            nuflux.knee_reweighting_model = knee

        def flux(mc_p_energy,mc_p_type,mc_p_zenith):
            mc_p_type = np.int32(mc_p_type)
            mc_p_type = conv.ConvertPrimaryFromPDG(mc_p_type)
            try:
                return fluxconst*nuflux.getFlux(mc_p_type,mc_p_energy,mc_p_zenith)
            except RuntimeError:
                if conv.IsPDGEncoded(mc_p_type,neutrino=True):
                    mc_p_type = conv.ConvertPrimaryFromPDG(mc_p_type)
                else:
                    mc_p_type = conv.ConvertPrimaryToPDG(mc_p_type)
                # FIXME: is this still the case?
                # the fluxes given by newnuflux are only for anti/neutrinos
                # so to calculate the total flux, there is a number of 2 necessary
                return 2*fluxconst*nuflux.getFlux(mc_p_type,mc_p_energy,mc_p_zenith)
        return flux

##############################################


def AtmoWrap(*args,**kwargs):
    """
    Allows currying atmospheric flux functions
    for class interface
    Args:
        *args: passed through to AtmosphericNuFlux
        **kwargs: passed through to AtmosphericNuFlux

    Returns: AtmosphericNuFlux with applied arguments

    """
    def wrapper():
        return AtmosphericNuFlux(*args,**kwargs)
    return wrapper

##############################################


def PowerWrap(*args,**kwargs):
    """
    Allows currying PowerLawFlux for class interface

    Args:
        *args: applied to PowerLawFlux
        **kwargs: applied to PowerLawFlux

    Returns: PowerLawFlux with applied arguments

    """

    def wrapper():
        return PowerLawFlux(*args,**kwargs)
    return wrapper

##############################################

def generated_corsika_flux(ebinc,datasets):
    """
    Calculate the livetime of a number of given coriska datasets using the weighting moduel
    The calculation here means a comparison of the number of produced events per energy bin
    with the expected event yield from fluxes in nature. If necessary call home to the simprod db.
    Works for 5C datasets.

    Args:
        ebinc (np.array): Energy bins (centers)
        datasets (list): A list of dictionaries with properties of the datasets or dataset numbers. If only nu8mbers are given, then simprod db will be queried
            format of dataset dict:
            example_datasets ={42: {"nevents": 1,\
                   "nfiles": 1,\
                   "emin": 1,\
                   "emax": 1,\
                   "normalization": [10., 5., 3., 2., 1.],\
                   "gamma": [-2.]*5,\
                   "LowerCutoffType": 'EnergyPerNucleon',\
                   "UpperCutoffType": 'EnergyPerParticle',\
                   "height": 1600,\
                   "radius": 800}}

    Returns:
        tuple (generated protons, generated irons)
    """
    
    if isinstance(datasets,dict):
        pass

    elif not isinstance(datasets,list):
        datasets = list(datasets)

    generators = []
    for ds in datasets:
       
        if not isinstance(datasets,dict):
            assert len(list(ds.values())) == 1, "Too many arguments per dataset"

        if isinstance(ds,int):
            db_result = from_simprod(ds)
            if isinstance(db_result, tuple):
                db_result = db_result[1]
            generators.append(db_result*datasets[ds])

        elif isinstance(list(ds.values())[0],int):
            db_result = from_simprod(int(list(ds.keys())[0]))
            if isinstance(db_result,tuple):
                db_result = db_result[1]
            generators.append(db_result*list(ds.values())[0])
        elif isinstance(list(ds.values())[0],dict):
            nfiles = ds.pop("nfiles")
            generators.append(nfiles*FiveComponent(**ds))
        else:
            raise ValueError("Problems understanding dataset properties {}".format(ds.__repr__()))

    gensum = reduce(lambda x, y: x + y, generators)
    p_gen  = AREA_SUM*gensum(ebinc,2212)
    fe_gen = AREA_SUM*gensum(ebinc,1000260560)

    return p_gen, fe_gen

##################################################################


class NuFluxes(object):
    """
    Namespace for neutrino fluxes
    """
    Honda2006    = staticmethod(AtmosphericNuFlux("honda2006"))
    Honda2006H3a = staticmethod(AtmosphericNuFlux("honda2006",knee="gaisserH3a_elbert")) 
    Honda2006H4a = staticmethod(AtmosphericNuFlux("honda2006",knee="gaisserH4a_elbert"))
    ERS          = staticmethod(AtmosphericNuFlux("sarcevic_std"))
    ERSH3a       = staticmethod(AtmosphericNuFlux("sarcevic_std",knee="gaisserH3a_elbert"))
    ERSH4a       = staticmethod(AtmosphericNuFlux("sarcevic_std",knee="gaisserH4a_elbert"))
    BERSSH3a     = staticmethod(AtmosphericNuFlux("BERSS_H3a_central"))
    BERSSH4a     = staticmethod(AtmosphericNuFlux("BERSS_H3p_central"))
    E2           = staticmethod(PowerLawFlux())
    BARTOL       = staticmethod(AtmosphericNuFlux("bartol"))


class MuFluxes(object):
    """
    Namespace for atmospheric muon fluxes
    """
    GaisserH3a   = staticmethod(ICMuFluxes.GaisserH3a)
    GaisserH4a   = staticmethod(ICMuFluxes.GaisserH4a)
    Hoerandel    = staticmethod(ICMuFluxes.Hoerandel)
    Hoerandel5   = staticmethod(ICMuFluxes.Hoerandel5)

    # if these fluxes are available depends on the version
    # of the weighting module
    
    #HoerandelIT  = staticmethod(mufluxes.Hoerandel_IT)
    #TIG1996      = staticmethod(mufluxes.TIG1996)
    #GlobalFitGST = staticmethod(mufluxes.GlobalFitGST)
    #Honda2004    = staticmethod(mufluxes.Honda2004)



"""
Provides containers for in-memory variable. These containers are called "categroies",
and they represent a set of variables for a certain type of data. Categories can
be further grouped into "Datasets". Variables can be read out from files and stored
in memory in the form of numpy arrays or pandas DataSeries/DataFrames. Selection criteria
can be applied simultaniously (and reversibly) to all categories in a dataset with the "Cut"
class.

HErmes.selection provides the following submodules:

- `categories` : Container classes for variables.

- `dataset` : Grouping categories together.

- `cut` : Apply selection criteria on variables in a category.

- `variables` : Variable definition. Harvest variables from files.

- `magic_keywords` : A bunch of fixed names for automatic weight calculation.


"""

import hjson
import os
import os.path
import inspect
import importlib
import re
import numpy as np

from ..utils import Logger

from . import categories as c
from . import dataset as ds
from ..icecube_goodies import weighting as wgt
from ..analysis import fluxes as fluxes

def load_dataset(config,
                 variables=None,
                 max_cpu_cores=c.MAX_CORES,
                 only_nfiles=None,
                 dtype=np.float64):
    """
    Read a json configuration file and load a dataset populated
    with variables from the files given in the configuration file.

    Args:
        config   (str/dict): json style config file or dict

    Keyword Args:
        variables (list)   : list of strings of variable names to read out
        max_cpu_cores (int): maximum number of cpu ucores to use for variable readout
        only_nfiles (int)  : readout only 'only_nfiles' 
        dtype (np.dtype)   : cast to the given datatype. By default it will be always double
                             (which is np.float64), however ofthen times it is advisable
                             to downcast to a less precise type to save memory.
    Returns:
        HErmes.selection.dataset.Dataset

    """
    cfg = config
    if not isinstance(config, dict):
        assert os.path.exists(config), "Config file {} does not exist!".format(config)
        cfg = hjson.load(open(config))

    categories = dict()
    weightfunctions = dict()
    models = dict()
    model_args = dict()
    files_basepath   = cfg["files_basepath"]
    for cat in list(cfg["categories"].keys()):
        thiscat = cfg["categories"][cat]
        sanitizer = lambda x: True

        if "file_regex" in thiscat:
            to_sanitize = thiscat["file_regex"]
            pattern = re.compile(to_sanitize)
            Logger.debug("Will look for files with pattern {}".format(pattern)) 
            def sanitizer(x):
                result = pattern.search(x)
                if result is None:
                    return False
                else:
                    return True
                #if to_sanitize in x:
                #    return True
                #else: 
                #    return False

        if thiscat["datatype"] == "simulation":
            categories[cat] = c.Simulation(cat)
            # remember that json keys are strings, so 
            # convert to int
            datasets = {}
            if "datasets" in thiscat:
                #datasets = {int(x): int(thiscat['datasets'][x]) for x in thiscat['datasets']}
                datasets = {x: int(thiscat['datasets'][x]) for x in thiscat['datasets']}
                
            
            categories[cat].get_files(os.path.join(files_basepath,\
                                                   thiscat['subpath']),\
                                                   prefix=thiscat["file_prefix"],\
                                                   datasets=datasets,\
                                                   sanitizer=sanitizer,\
                                                   only_nfiles=only_nfiles,\
                                                   ending=thiscat["file_type"])

            #weightfunctions[cat] = dict(inspect.getmembers(wgt))[thiscat["model_method"]]
            if not "model" in thiscat:
                models[cat] = None
                model_args[cat] = [None]
            else:
                try:
                    fluxclass, flux = thiscat["model"].split(".")
                    #fluxclass = thiscat["model"]
                    #flux = "__call__"
                    models[cat] = getattr(dict(inspect.getmembers(fluxes))[fluxclass],flux)
                    model_args[cat] = thiscat["model_args"]
                except ValueError:
                    Logger.warning("{} does not seem to be a valid model for {}. This might cause troubles. If not, it is probably fine!".format(thiscat["model"],cat))
                    models[cat] = None

        elif thiscat["datatype"] == "data":
            categories[cat] = c.Data(cat)
            categories[cat].get_files(os.path.join(files_basepath,thiscat['subpath']),\
                                      prefix=thiscat["file_prefix"],\
                                      sanitizer=sanitizer,\
                                      ending=thiscat["file_type"])
            #models[cat] = float(thiscat["livetime"])
            #weightfunctions[cat] = dict(inspect.getmembers(wgt))[thiscat["model_method"]]
            
        elif thiscat["datatype"] == "reweighted":
            pass
        else:
            raise TypeError("Data type not understood. Has to be either 'simulation', 'reweighted' or 'data'!!")

    # at last we can take care of reweighted categories
    for cat in list(cfg["categories"].keys()):
        thiscat = cfg["categories"][cat]
        if thiscat["datatype"] == "reweighted":
            categories[cat] = c.ReweightedSimulation(cat,categories[thiscat["parent"]])
            #if thiscat["model"]:
            #    fluxclass, flux = thiscat["model"].split(".")
            #    models[cat] = getattr(dict(inspect.getmembers(fluxes))[fluxclass],flux)
            #    weightfunctions[cat] = dict(inspect.getmembers(wgt))[thiscat["model_method"]]
        elif thiscat["datatype"] in ["data", "simulation"]:
            pass
        else:
            raise TypeError("Data type not understood. Has to be either 'simulation', 'reweighted' or 'data'!!")

    for cat in categories:
        if isinstance(categories[cat], c.Data):
            continue
        if not "weights" in  cfg["categories"][cat]:
            continue
        categories[cat].weightvarname = cfg["categories"][cat]["weights"]

    #combined categories
    combined_categories = dict() 
    for k in list(combined_categories.keys()):
        combined_categories[k] = [categories[l] for l in cfg["combined_categories"]]

    # import variable defs
    vardefs = cfg["variable_definitions"]
    if isinstance(vardefs, str) or isinstance(vardefs, unicode):
        vardefs = importlib.import_module(cfg["variable_definitions"])
    elif isinstance(vardefs, dict):
        vardefs = {}
        for k in cfg["variable_definitions"]:
            vardefs[k] = importlib.import_module(cfg["variable_definitions"][k])
    else:
        raise ValueError("Can not understand variable definitions {} of type {}".\
                         format(vardefs, type(vardefs)))
    #vardefs = importlib.import_module(cfg["variable_definitions"])
    dataset = ds.Dataset(*list(categories.values()),\
                         combined_categories=combined_categories)

    dataset.load_vardefs(vardefs)
    dataset.read_variables(names=variables, max_cpu_cores=max_cpu_cores, dtype=dtype)
    #dataset.set_weightfunction(weightfunctions)
    #dataset.get_weights(models=models)
    dataset.calculate_weights(model=models, model_args=model_args)
    plot_dict = {}
    for k in cfg["categories"]:
        if "plotting" in cfg["categories"][k]:
            plot_dict[k] = cfg["categories"][k]["plotting"]
    dataset.set_default_plotstyles(plot_dict)
    return dataset



"""
Datasets group categories together. Method calls on datasets invoke the individual methods
on the individual categories. Cuts applied to datasets will act on each individual category.

"""

import pandas as pd
import numpy as np

from collections import OrderedDict
from copy import deepcopy as copy

from ..visual import VariableDistributionPlot
from ..utils import isnotebook
from ..utils import Logger
from dashi.tinytable import TinyTable

from . import categories

def get_label(category):
    """
    Get the label for labeling plots from a datasets plot_options dictionary.

    Args:
        category (HErmes.selection.categories.category): Query the category's plot_options dict, if not fall back to category.name

    Returns:
        string
    """

    if category.plot_options:
        if "label" in category.plot_options:
            return category.plot_options["label"]
        else:
            return category.name
    else:
        return category.name


class Dataset(object):
    """
    Holds different categories, relays calls to each
    of them.
    """

    def __init__(self, *args, **kwargs):
        """
        Args:
            *args: HErmes.selection.variables.categories.Category list

        Keyword Args:
            combined_categories:
        """
        self.categories = []
        self.combined_categories = []
        # sort categories, do reweighted simulation last
        # FIXME: if not, there will be problems
        # FIXME: investigate!
        reweighted_categories = []

        self.default_plotstyles = {}
        for cat in args:
            self.__dict__[cat.name] = cat
            if isinstance(cat,categories.ReweightedSimulation):
                reweighted_categories.append(cat)
                continue 
            self.categories.append(cat)
        self.categories = self.categories + reweighted_categories
        if 'combined_categories' in kwargs:
            for name in list(kwargs['combined_categories'].keys()):
                self.combined_categories.append(categories.CombinedCategory(name,kwargs['combined_categories'][name]))

    def set_default_plotstyles(self, styledict):
        """
        Define a standard for each category how 
        it should appear in plots

        Args:
            styledict (dict)
        """
        self.default_plotstyles = styledict
        for cat in self.categorynames:
            self[cat].add_plotoptions(styledict[cat])

    def add_variable(self, variable):
        """
        Add a variable to this category

        Args:
            variable (HErmes.selection.variables.variables.Variable): A Variable instalce
        """

        for cat in self.categories:
            cat.add_variable(variable)

    def delete_variable(self, varname):
        """
        Delete a variable entirely from the dataset

        Args:
            varname (str): the name of the variable

        Returns:
            None
        """
        for cat in self.categories:
            cat.delete_variable(varname)

    def load_vardefs(self, vardefs):
        """
        Load the variable definitions from a module

        Args:
            vardefs (python module/dict): A module needs to contain variable definitions.
                                         It can also be a dictionary of categoryname->module

        """
        if isinstance(vardefs, dict):
            for k in vardefs:
                # FIXME: the way over self.__dict__ does not work
                # maybe there is something more fishy...
                if str(k) == "all":
                    for cat in self.categories:
                        cat.load_vardefs(vardefs[k])

                for cat in self.categories:
                    if cat.name == k:
                        cat.load_vardefs(vardefs[k])
                #self.__dict__[k].load_vardefs(vardefs)

        else:
            for cat in self.categories:
                cat.load_vardefs(vardefs)

    @property
    def variablenames(self):
        return {cat.name : cat.variablenames for cat in self.categories}

    @property
    def files(self):
        return {cat.name : cat.files for cat in self.categories}

    #@GetTiming
    def read_variables(self, names=None,
                             max_cpu_cores=categories.MAX_CORES,
                             dtype=np.float64):
        """
        Read out the variable for all categories

        Keyword Args:
            names (str): Readout only these variables if given
            max_cpu_cores (int): Maximum number of cpu cores which will be used
            dtype (np.dtype) : Cast to the given datatype (default is np.flaot64)
        Returns:
            None
        """
        progbar = False
        try:
            import tqdm
            n_it = len(self.categories)
            loader_string = "Loading dataset"
            if isnotebook():
                bar = tqdm.tqdm_notebook(total=n_it, desc=loader_string, leave=True)
            else:
                bar = tqdm.tqdm(total=n_it, desc=loader_string, leave=True)
            progbar = True
        except ImportError:
            pass
        for cat in self.categories:
            Logger.debug("Reading variables for {}".format(cat))
            cat.read_variables(names=names, max_cpu_cores=max_cpu_cores, dtype=dtype)
            if progbar: bar.update()

    def drop_empty_variables(self):
        """
        Delete variables which have no len

        Returns:
            None
        """

        for cat in self.categories:
            cat.drop_empty_variables()

    def set_weightfunction(self, weightfunction=lambda x:x):
        """
        Defines a function which is used for weighting

        Args:
            weightfunction (func or dict): if func is provided, set this to all categories
                                           if needed, provide dict, cat.name -> func for individula setting

        Returns:
            None
        """
        if isinstance(weightfunction, dict):
            for cat in self.categories:
                cat.set_weightfunction(weightfunction[cat.name])

        else:
            for cat in self.categories:
                cat.set_weightfunction(weightfunction)

    def calculate_weights(self, model=None, model_args=None):
        """
        Calculate the weights for all categories

        Keyword Args:
            model (dict/func) : Either a dict catname -> func or a single func
                                 If it is a single funct it will be applied to all categories
            model_args (dict/list): variable names as arguments for the function
        """
        if isinstance(model, dict):
            if not isinstance(model_args, dict):
                raise ValueError("if model is a dict, model_args has to be a dict too!")
            for catname in model:
                self.get_category(catname).calculate_weights(model=model[catname], model_args=model_args[catname])

        else:
            for cat in self.categories:
                cat.calculate_weights(model=model, model_args=model_args)

    # def get_weights(self, models):
    #     """
    #     Calculate the weights for all categories
    #
    #     Args:
    #         models (dict or callable): A dictionary of categoryname -> model or a single clbl
    #     """
    #     if isinstance(models, dict):
    #         for catname in models:
    #             self.get_category(catname).get_weights(models[catname])
    #     if callable(models):
    #         for cat in self.categories:
    #             cat.get_weights(models)

    def add_category(self,category):
        """
        Add another category to the dataset

        Args:
            category (HErmes.selection.categories.Category): add this category

        """

        self.categories.append(category)

    def __getitem__(self, item):
        """
        Shortcut for self.get_category/get_variable

        Args:
            item:

        Returns:
            HErmes.selection.variables.Variable/HErmes.selection.categories.Category
        """
        try:
            return self.get_category(item)
        except KeyError:
            pass
        try:
            return self.get_variable(item)
        except  KeyError:
            pass
        if ":" in item:
            cat, var = item.split(":")
            return self.get_category(cat).get(var)

        else:
            raise KeyError("{} can not be found".format(item))

    def get_category(self, categoryname):
        """
        Get a reference to a category.

        Args:
            category: A name which has to be associated to a category

        Returns:
             HErmes.selection.categories.Category
        """

        for cat in self.categories:
            if cat.name == categoryname:
                return cat

        raise KeyError("Can not find category {}.".format(categoryname))

    def get_variable(self, varname):
        """
        Get a pandas dataframe for all categories

        Args:
            varname (str): A name of a variable

        Returns:
            pandas.DataFrame: A 2d dataframe category -> variable
        """

        var = dict()
        for cat in self.categories:
            var[cat.name] = cat.get(varname)

        return pd.DataFrame.from_dict(var, orient="index")

    def set_livetime(self, livetime):
        """
        Define a livetime for this dataset.

        Args:
            livetime (float): Time interval the data was taken in. (Used for rate calculation)

        Returns:
            None
        """
        for cat in self.categories:
            if hasattr(cat, "set_livetime"):
                cat.set_livetime(livetime)

    @property
    def weights(self):
        """
        Get the weights for all categories in this dataset
        """
        w = dict()
        for cat in self.categories:
            w[cat.name] = pd.Series(cat.weights, dtype=np.float64)
        return pd.DataFrame.from_dict(w,orient='index')

    def __repr__(self):
        """
        String representation
        """

        rep = """ <Dataset: """
        for cat in self.categories:
            rep += "{} ".format(cat.name)
        rep += ">"
        return rep

    def add_cut(self,cut):
        """
        Add a cut without applying it yet

        Args:
            cut (HErmes.selection.variables.cut.Cut): Append this cut to the internal cutlist
        """
        for cat in self.categories:
            cat.add_cut(cut)

    def apply_cuts(self,inplace=False):
        """
        Apply them all!
        """
        for cat in self.categories:
            cat.apply_cuts(inplace=inplace)

    def undo_cuts(self):
        """
        Undo previously done cuts, but keep them so that
        they can be re-applied
        """
        for cat in self.categories:
            cat.undo_cuts()

    def delete_cuts(self):
        """
        Completely purge all cuts from this
        dataset
        """
        for cat in self.categories:
            cat.delete_cuts()

    @property
    def categorynames(self):
        return [cat.name for cat in self.categories]

    @property
    def combined_categorynames(self):
        return [cat.name for cat in self.combined_categories]

    def get_sparsest_category(self, omit_empty_cat=True):
        """
        Find out which category of the dataset has the least statistical power

        Keyword Args:
            omit_empty_cat (bool): if a category has no entries at all, omit
        Returns:
            str: category name
        """

        name  = self.categories[0].name
        count = self.categories[0].raw_count
        for cat in self.categories:
            if cat.raw_count < count:
                if (cat.raw_count == 0) and omit_empty_cat:
                    continue
                count = cat.raw_count
                name  = cat.name
        return name


    def distribution(self,name,\
                     ratio=([],[]),
                     cumulative=True,
                     log=False,
                     transform=None,
                     disable_weights=False,
                     color_palette='dark',
                     normalized = False,
                     styles = dict(),
                     style="classic",
                     ylabel="rate/bin [1/s]",
                     axis_properties=None,
                     ratiolabel="data/$\Sigma$ bg",
                     bins=None,
                     external_weights=None,
                     savepath=None,
                     figure_factory=None,
                     zoomin=False,
                     adjust_ticks = lambda x  : x):
        """
        One shot short-cut for one of the most used
        plots in eventselections.

        Args:
            name                 (string) : The name of the variable to plot

        Keyword Args:
            path                    (str) : The path under which the plot will be saved.
            ratio                  (list) : A ratio plot of these categories will be crated
            color_palette           (str) : A predifined color palette (from seaborn or HErmes.plotting.colors) 
            normalized             (bool) : Normalize the histogram by number of events
            transform          (callable) : Apply this transformation before plotting
            disable_weights        (bool) : Disable all weighting to avoid problems with uneven sized arrays
            styles                 (dict) : plot styling options
            ylabel                  (str) : general label for y-axis
            ratiolabel              (str) : different label for the ratio part of the plot
            bins             (np.ndarray) : binning, if None binning will be deduced from the variable definition
            figure_factory         (func) : factory function which return a matplotlib.Figure
            style                (string) : TODO "modern" || "classic" || "modern-cumul" || "classic-cumul"
            savepath             (string) : Save the canvas at given path. None means it will not be saved.
            external_weights       (dict) : supply external weights - this will OVERIDE ANY INTERNALLY CALCULATED WEIGHTS
                                            and use the supplied weights instead.
                                            Must be in the form { "categoryname" : weights}
            axis_properties        (dict) : Manually define a plot layout with up to three axes.
                                            For example, it can look like this:
                                            {
                                                "top": {"type": "h", # histogram
                                                        "height": 0.4, # height in percent
                                                        "index": 2}, # used internally
                                                "center": {"type": "r", # ratio plot
                                                            "height": 0.2,
                                                            "index": 1},
                                                "bottom": { "type": "c", # cumulative histogram
                                                            "height": 0.2,
                                                            "index": 0}
                                            }

            zoomin                 (bool) : If True, select the yrange in a way that the interesting part of the 
                                            histogram is shown. Caution is needed, since this might lead to an
                                            overinterpretation of fluctuations.
            adjust_ticks            (fcn) : A function, applied on a matplotlib axes
                                           which will set the proper axis ticks
        Returns:
            HErmes.selection.variables.VariableDistributionPlot
        """
        
        
        # if (not cumulative) or ratio  == ([],[]):
        #
        #     # assuming a single cumulative axis
        #     tmp_axis_properties = dict()
        #     unassigned_height = 0
        #
        #     for key in axis_properties:
        #         if ("c" == axis_properties[key]["type"]) and (not cumulative):
        #             unassigned_height += axis_properties[key]["height"]
        #             continue
        #         if ("r" == axis_properties[key]["type"]) and (ratio == ([],[])):
        #             unassigned_height += axis_properties[key]["height"]
        #             continue
        #
        #         tmpdict = copy(axis_properties[key])
        #         tmpdict["index"] = tmpdict["index"] -1 - bool(ratio == ([],[]))
        #         tmp_axis_properties.update({key : tmpdict})
        #
        #     n_plots = len(tmp_axis_properties.keys())
        #     extra_height = unassigned_height/float(n_plots)
        #     for key in tmp_axis_properties:
        #         tmp_axis_properties[key]["height"] += extra_height
        #
        # else:
        #     tmp_axis_properties = copy(axis_properties)

        if axis_properties is not None:
            tmp_axis_properties = copy(axis_properties)

        else:
            # always have the histogram, but add
            # cumulative or ratio plot
            if cumulative and ratio != ([],[]):
                tmp_axis_properties = {\
                    "top": {"type": "h", \
                            "height": 0.4, \
                            "index": 2},\
                    "center": {"type": "r",\
                               "height": 0.2,\
                               "index": 1},\
                    "bottom": {"type": "c", \
                               "height": 0.2,\
                               "index": 0}\
                    }
            elif cumulative:
                tmp_axis_properties = { \
                    "top": {"type": "h", \
                            "height": 0.6, \
                            "index": 1}, \
                    "bottom": {"type": "c", \
                               "height": 0.4, \
                               "index": 0} \
                    }
            elif ratio != ([],[]):
                tmp_axis_properties = { \
                    "top": {"type": "h", \
                            "height": 0.6, \
                            "index": 1}, \
                    "bottom": {"type": "r", \
                               "height": 0.4, \
                               "index": 0} \
                    }
            else:
                tmp_axis_properties = { \
                    "top": {"type": "h", \
                            "height": 0.95, \
                            "index": 0}, \
                    }
        axes_locator = [(tmp_axis_properties[k]["index"], tmp_axis_properties[k]["type"], tmp_axis_properties[k]["height"])\
                        for k in tmp_axis_properties]
        #print (axes_locator)
        #heights = [axis_properties[k]["height"] for k in axis_properties]
        cuts = self.categories[0].cuts
        sparsest = self.get_sparsest_category()

        # check if there are user-defined bins for that variable
        if bins is None:
            bins = self.get_category(sparsest).vardict[name].bins
        # calculate the best possible binning
        if bins is None:
            bins = self.get_category(sparsest).vardict[name].calculate_fd_bins()
        label = self.get_category(sparsest).vardict[name].label
        plot = VariableDistributionPlot(cuts=cuts, bins=bins,\
                                        xlabel=label,\
                                        color_palette=color_palette)
        if styles:
            plot.plot_options = styles
        else:
            plot.plot_options = self.default_plotstyles
        plotcategories = self.categories + self.combined_categories 

        Logger.warning("For variables with different lengths the weighting is broken. If weights, it will fail")
        for cat in [x for x in plotcategories if x.plot]:
            if external_weights is not None:
                weights = external_weights[cat.name]
            elif ((cat.weights is not None) and (not disable_weights)):
                weights = cat.weights
                Logger.debug(f"Found {len(weights)} weights")
                if not len(weights):
                    weights = None
            else:
                weights = None
            Logger.debug(f"Adding variable data {name}")
            plot.add_variable(cat, name, transform=transform, external_weights=weights)
            if cumulative:
                Logger.debug("Adding variable data {} for cumulative plot".format(name))
                plot.add_cumul(cat.name)

        if len(ratio[0]) and len(ratio[1]):
            Logger.debug("Requested to plot ratio {} {}".format(ratio[0], ratio[1]))
            tratio,tratio_err = self.calc_ratio(nominator=ratio[0],\
                                                denominator=ratio[1])

            plot.add_ratio(ratio[0],ratio[1],\
                           total_ratio=tratio,\
                           label=ratiolabel,
                           total_ratio_errors=tratio_err)

        plot.plot(axes_locator=axes_locator,\
                  normalized=normalized,\
                  figure_factory=figure_factory,\
                  log=log,\
                  style=style,\
                  ylabel=ylabel,\
                  zoomin=zoomin,\
                  adjust_ticks=adjust_ticks)
        #plot.add_legend()
        #plot.canvas.save(savepath,savename,dpi=350)
        if savepath is not None:
            plot.canvas.save(savepath, name)
        return plot

    @property
    def integrated_rate(self):
        """
        Integrated rate for each category

        Returns:
            pandas.Panel: rate with error
        """

        rdata,edata,index = [],[],[]
        for cat in self.categories + self.combined_categories:
            rate,error = cat.integrated_rate
            rdata.append(rate)
            index.append(cat.name)
            edata.append(error)

        rate = pd.Series(rdata,index)
        err  = pd.Series(edata,index)
        return (rate,err)

    #FIXME static method!
    def sum_rate(self,categories=None):
        """
        Sum up the integrated rates for categories

        Args:
            categories: categories considerred background

        Returns:
             tuple: rate with error

        """
        if categories is None:
            return 0,0

        categories = [self.get_category(i) if isinstance(i, str) else i for i in categories]
        rate,error = categories[0].integrated_rate
        error = error**2
        for cat in categories[1:]:
            tmprate,tmperror = cat.integrated_rate
            rate  += tmprate # categories should be independent
            error += tmperror**2
        return (rate,np.sqrt(error))

    def calc_ratio(self,nominator=None,denominator=None):
        """
        Calculate a ratio of the given categories

        Args:
            nominator (list):
            denominator (list):

        Returns:
            tuple
        """
        nominator = [self.get_category(i) if isinstance(i, str) else i for i in nominator]
        denominator = [self.get_category(i) if isinstance(i, str) else i for i in denominator]

        a,a_err = self.sum_rate(categories=nominator)
        b,b_err = self.sum_rate(categories=denominator)
        if b == 0:
            return np.nan, np.nan
        sum_err = np.sqrt((a_err/ b) ** 2 + ((-a * b_err)/ (b ** 2)) ** 2)
        return a/b, sum_err

    def _setup_table_data(self,signal=None,background=None):
        """
        Setup data for a table
        If signal and background are given, also summed values
        will be in the list

        Keyword Args:
            signal (list): category names which are considered signal
            background (list): category names which are considered background

        Returns
            dict: table dictionary
        """

        rates, errors = self.integrated_rate
        sgrate, sgerrors = self.sum_rate(signal)
        bgrate, bgerrors = self.sum_rate(background)
        allrate, allerrors = self.sum_rate(self.categories)
        tmprates  = pd.Series([sgrate,bgrate,allrate],index=["signal","background","all"])
        tmperrors = pd.Series([sgerrors,bgerrors,allerrors],index=["signal","background","all"])
        rates = rates.append(tmprates)
        errors = errors.append(tmperrors)

        datacats = []
        for cat in self.categories + self.combined_categories:
            if isinstance(cat,categories.Data):
                datacats.append(cat)
        if datacats:
            simcats = [cat for cat in self.categories if cat.name not in [kitty.name for kitty in datacats]]
            simrate, simerror = self.sum_rate(simcats)

        fudges = dict()
        for cat in datacats:
            rate,error = cat.integrated_rate
            try:
                fudges[cat.name] = (rate/simrate),(error/simerror)
            except ZeroDivisionError:
                fudges[cat.name] = np.NaN
        rate_dict = OrderedDict()
        all_fudge_dict = OrderedDict()
        #for catname in sorted(self.categorynames) + sorted(self.combined_categorynames):
        for cat in datacats:
            label = get_label(cat)
            #cfg = GetCategoryConfig(cat.name)
            #label = cfg["label"]
            rate_dict[label] = (rates[cat.name], errors[cat.name])
            if cat.name in fudges:
                all_fudge_dict[label] = fudges[cat.name]
            else:
                all_fudge_dict[label] = None

        rate_dict["Sig."] =  (rates["signal"],errors["signal"] )
        rate_dict["Bg."] = (rates["background"],errors["background"])
        rate_dict["Gr. Tot."] = (rates["all"],errors["all"])
        all_fudge_dict["Sig."]     = None
        all_fudge_dict["Bg."]      = None
        all_fudge_dict["Gr. Tot."] = None
        return rate_dict,all_fudge_dict

    def tinytable(self,signal=None,\
                    background=None,\
                    layout="v",\
                    format="html",\
                    order_by=lambda x:x,
                    livetime=1.):
        """
        Use dashi.tinytable.TinyTable to render a nice
        html representation of a rate table

        Args:
            signal (list) : summing up signal categories to calculate total signal rate
            background (list): summing up background categories to calculate total background rate
            layout (str) : "v" for vertical, "h" for horizontal
            format (str) : "html","latex","wiki"

        Returns:
            str: formatted table in desired markup
        """
        def cellformatter(input):
            #print input
            if input is None:
                return "-"
            if isinstance(input[0],pd.Series):
                input = (input[1][0],input[1][0])
            return "{:4.2e} +- {:4.2e}".format(input[0],input[1])

        #FIXME: sort the table columns
        rates,fudges = self._setup_table_data(signal=signal,background=background)
        events = dict()
        for k in rates:
            events[k] = rates[k][0] * livetime, rates[k][1] * livetime

        showcats =  [get_label(cat) for cat in self.categories if cat.show_in_table]
        showcats += [get_label(cat) for cat in self.combined_categories if cat.show_in_table]
        showcats.extend(['Sig.',"Bg.","Gr. Tot."])
        orates = OrderedDict()
        ofudges = OrderedDict()
        oevents = OrderedDict()
        for k in list(rates.keys()):
            if k in showcats:
                orates[k] = rates[k]
                ofudges[k] = fudges[k]
                oevents[k] = events[k]
        #rates  = {k : rates[k] for k in rates if k in showcats}
        #fudges = {k : fudges[k] for k in fudges if k in showcats}
        #events = {k : events[k] for k in events if k in showcats}
        tt = TinyTable()

        #bypass the add function ot add an ordered dict
        for label,data in [('Rate (1/s)', orates),("Ratio", ofudges),("Events",oevents)]:
            tt.x_labels.append(label)
            tt.label_data[label] = data
        #tt.add("Rate (1/s)", **rates)
        #tt.add("Ratio",**fudges)
        #tt.add("Events",**events)
        return tt.render(layout=layout,format=format,\
                         format_cell=cellformatter,\
                         order_by=order_by)


    #def cut_progression_table(self,cuts,\
    #                signal=None,\
    #                background=None,\
    #                layout="v",\
    #                format="html",\
    #                order_by=lambda x:x,
    #                livetime=1.):

    #    self.delete_cuts()
    #    self.undo_cuts()
    #    for cut in cuts:
    #        self.add_cut(cut)
    #        self.apply_cuts()

    def __len__(self):
        #FIXME: to be implemented
        raise NotImplementedError



"""
Unit conversions and such
"""

from numpy import vectorize,int32
#from icecube import icetray,dataclasses,NewNuFlux
####################################################


#FIXME: boilerplate this here, because it is not possible
# to load two versions of icerec simultaniously
# This should go into icerec!

class ParticleType(object):
    """
    Namespace for icecube particle type codes
    """
    unknown = 0
    Gamma = 1
    EPlus = 2
    EMinus = 3
    MuPlus = 5
    MuMinus = 6
    Pi0 = 7
    PiPlus = 8
    PiMinus = 9
    K0_Long = 10
    KPlus = 11
    KMinus = 12
    Neutron = 13
    PPlus = 14
    PMinus = 15
    K0_Short = 16
    NuE = 66
    NuEBar = 67
    NuMu = 68
    NuMuBar = 69
    TauPlus = 131
    TauMinus = 132
    NuTau = 133
    NuTauBar = 134
    He4Nucleus = 402
    Li7Nucleus = 703
    Be9Nucleus = 904
    B11Nucleus = 1105
    C12Nucleus = 1206
    N14Nucleus = 1407
    O16Nucleus = 1608
    F19Nucleus = 1909
    Ne20Nucleus = 2010
    Na23Nucleus = 2311
    Mg24Nucleus = 2412
    Al26Nucleus = 2613
    Al27Nucleus = 2713
    Si28Nucleus = 2814
    P31Nucleus = 3115
    S32Nucleus = 3216
    Cl35Nucleus = 3517
    Ar36Nucleus = 3618
    Ar37Nucleus = 3718
    Ar38Nucleus = 3818
    Ar39Nucleus = 3918
    Ar40Nucleus = 4018
    Ar41Nucleus = 4118
    Ar42Nucleus = 4118
    K39Nucleus = 3919
    Ca40Nucleus = 4020
    Sc45Nucleus = 4521
    Ti48Nucleus = 4822
    V51Nucleus = 5123
    Cr52Nucleus = 5224
    Mn55Nucleus = 5525
    Fe56Nucleus = 5626

#######################################################

class PDGCode(object):
    """
    Namespace for PDG conform particle type codes
    """
    unknown = 0
    Gamma = 22
    EPlus = -11
    EMinus = 11
    MuPlus = -13
    MuMinus = 13
    Pi0 = 111
    PiPlus = 211
    PiMinus = -211
    K0_Long = 130
    KPlus = 321
    KMinus = -321
    Neutron = 2112
    PPlus = 2212
    PMinus = -2212
    K0_Short = 310
    Eta = 221
    Lambda = 3122
    SigmaPlus = 3222
    Sigma0 = 3212
    SigmaMinus = 3112
    Xi0 = 3322
    XiMinus = 3312
    OmegaMinus = 3334
    NeutronBar = -2112
    LambdaBar = -3122
    SigmaMinusBar = -3222
    Sigma0Bar = -3212
    SigmaPlusBar = -3112
    Xi0Bar = -3322
    XiPlusBar = -3312
    OmegaPlusBar = -3334
    DPlus = 411
    DMinus = -411
    D0 = 421
    D0Bar = -421
    DsPlus = 431
    DsMinusBar = -431
    LambdacPlus = 4122
    WPlus = 24
    WMinus = -24
    Z0 = 23
    NuE = 12
    NuEBar = -12
    NuMu = 14
    NuMuBar = -14
    TauPlus = -15
    TauMinus = 15
    NuTau = 16
    NuTauBar = -16
    He3Nucleus = 1000020030
    He4Nucleus = 1000020040
    Li6Nucleus = 1000030060
    Li7Nucleus = 1000030070
    Be9Nucleus = 1000040090
    B10Nucleus = 1000050100
    B11Nucleus = 1000050110
    C12Nucleus = 1000060120
    C13Nucleus = 1000060130
    N14Nucleus = 1000070140
    N15Nucleus = 1000070150
    O16Nucleus = 1000080160
    O17Nucleus = 1000080170
    O18Nucleus = 1000080180
    F19Nucleus = 1000090190
    Ne20Nucleus = 1000100200
    Ne21Nucleus = 1000100210
    Ne22Nucleus = 1000100220
    Na23Nucleus = 1000110230
    Mg24Nucleus = 1000120240
    Mg25Nucleus = 1000120250
    Mg26Nucleus = 1000120260
    Al26Nucleus = 1000130260
    Al27Nucleus = 1000130270
    Si28Nucleus = 1000140280
    Si29Nucleus = 1000140290
    Si30Nucleus = 1000140300
    Si31Nucleus = 1000140310
    Si32Nucleus = 1000140320
    P31Nucleus = 1000150310
    P32Nucleus = 1000150320
    P33Nucleus = 1000150330
    S32Nucleus = 1000160320
    S33Nucleus = 1000160330
    S34Nucleus = 1000160340
    S35Nucleus = 1000160350
    S36Nucleus = 1000160360
    Cl35Nucleus = 1000170350
    Cl36Nucleus = 1000170360
    Cl37Nucleus = 1000170370
    Ar36Nucleus = 1000180360
    Ar37Nucleus = 1000180370
    Ar38Nucleus = 1000180380
    Ar39Nucleus = 1000180390
    Ar40Nucleus = 1000180400
    Ar41Nucleus = 1000180410
    Ar42Nucleus = 1000180420
    K39Nucleus = 1000190390
    K40Nucleus = 1000190400
    K41Nucleus = 1000190410
    Ca40Nucleus = 1000200400
    Ca41Nucleus = 1000200410
    Ca42Nucleus = 1000200420
    Ca43Nucleus = 1000200430
    Ca44Nucleus = 1000200440
    Ca45Nucleus = 1000200450
    Ca46Nucleus = 1000200460
    Ca47Nucleus = 1000200470
    Ca48Nucleus = 1000200480
    Sc44Nucleus = 1000210440
    Sc45Nucleus = 1000210450
    Sc46Nucleus = 1000210460
    Sc47Nucleus = 1000210470
    Sc48Nucleus = 1000210480
    Ti44Nucleus = 1000220440
    Ti45Nucleus = 1000220450
    Ti46Nucleus = 1000220460
    Ti47Nucleus = 1000220470
    Ti48Nucleus = 1000220480
    Ti49Nucleus = 1000220490
    Ti50Nucleus = 1000220500
    V48Nucleus = 1000230480
    V49Nucleus = 1000230490
    V50Nucleus = 1000230500
    V51Nucleus = 1000230510
    Cr50Nucleus = 1000240500
    Cr51Nucleus = 1000240510
    Cr52Nucleus = 1000240520
    Cr53Nucleus = 1000240530
    Cr54Nucleus = 1000240540
    Mn52Nucleus = 1000250520
    Mn53Nucleus = 1000250530
    Mn54Nucleus = 1000250540
    Mn55Nucleus = 1000250550
    Fe54Nucleus = 1000260540
    Fe55Nucleus = 1000260550
    Fe56Nucleus = 1000260560
    Fe57Nucleus = 1000260570
    Fe58Nucleus = 1000260580

#############################################################
ptype = ParticleType()
pdgtype = PDGCode()

# wrap these to dicts
#ptype_to_pdg = dict([(ptype.__getattribute__(x),pdgtype.__getattribute__(x))\
#                     for x in dir(ParticleType)\
#                     if not x.startswith('_') and x in dir(PDGCode)])
ptype_to_pdg = {ptype.__getattribute__(x) : pdgtype.__getattribute__(x)\
                     for x in dir(ParticleType)\
                     if not x.startswith('_') and x in dir(PDGCode)}
#pdg_to_ptype = dict([(pdgtype.__getattribute__(x),ptype.__getattribute__(x))\
#                     for x in dir(PDGCode)\
#                     if not x.startswith('_') and x in dir(ParticleType)])
pdg_to_ptype = {pdgtype.__getattribute__(x) : ptype.__getattribute__(x)\
                     for x in dir(PDGCode)\
                     if not x.startswith('_') and x in dir(ParticleType)}


def IsPDGEncoded(pid,neutrino=False):
    """
    Check if the particle has already a pdg compatible
    pid

    Args:
        id (int): Partilce Id

    Keyword Args:
        neutrino (bool): as nue is H in PDG, set true if you know already
                    that ihe particle might be a neutrino

    Returns (bool): True if PDG compatible
    """

    for i in pid.values:
        if i in list(pdg_to_ptype.keys()):
            if i == 14 and neutrino:
                return True
            elif i == 14 and not neutrino:
                return False

            else:
                return True

    return False

##############################


def ConvertPrimaryToPDG(pid):
    """
    Convert a primary id in an i3 file to the new values
    given by the pdg
    """
    def _convert(pid):
        if pid in ptype_to_pdg:
            return int32(ptype_to_pdg[pid])
        else:
            return int32(pid)

    return vectorize(_convert)(pid)

###############################


def ConvertPrimaryFromPDG(pid):
    """
    Convert a primary id in an i3 file to the new values
    given by the pdg
    """
    def _convert(pid):
        if pid in pdg_to_ptype:
            return int32(pdg_to_ptype[pid])
        else:
            return int32(pid)

    return vectorize(_convert)(pid)




"""
Models for particle fluxes. These are just examples, for specific cosmic ray modelss
have a look at e.g. https://github.com/afedynitch/CRFluxModels.git
"""
import numpy as np

class PowerLawFlux(object):
    """
    A flux only dependent on the energy of a particle, following a power law. Defined in
    an energy interval [emin, emax] with fluence phi0 and spectral index gamma
    """

    def __init__(self, emin, emax, phi0, gamma):
        """
        Args:
            emin  (float): minimum energy
            emax  (float): maximum energy
            phi0  (float): normalization
            gamma (float): spectral index
        """
        self.emin = emin
        self.emax = emax
        self.gamma = gamma
        self.phi0 = phi0

    @staticmethod
    def E2_1E8(energy):
        """
        A flux with fixed parameters, spectral index E**-2 and normalization 1E-8
        Usefull for automatic weighting.

        Args:
            energy
        """
        return 1e-8*np.power(energy, -2)

    def __call__(self, energy):
        """
        Calculate the flux for a given energy
        Args:
            energy (np.ndarray): primary energy

        Returns:
            np.ndarray
        """
        energy = energy.astype(np.float64)

        fl = self.phi0 * np.power(energy, self.gamma)
        fl[energy < self.emin] = 0.
        fl[energy > self.emax] = 0.
        return fl

    def fluxsum(self):
        """
        The integrated flux

        Returns:
            float
        """

        if self.gamma < -1:
            ex = 1 + self.gamma
            return self.phi0 * (self.emax ** ex - self.emin ** ex) / ex
        elif self.gamma == -1:
            return self.phi0 * (np.log(self.emax) - np.log(self.emin))
        else:
            raise ValueError("Integration of positive gamma not supported")

class Constant(object):

    @staticmethod
    def identity(x):
        return np.ones(len(x))



"""
Provide mathematical functions which can be used to create models. 
The functions have to be always in the form f(x, *parameters) where the
paramters will be fitted and x are the input values.
"""
import numpy as np
import scipy.stats as st

#import statsmodels.stats.gof as gof

from ..utils import Logger

def poisson(x, lmbda):
    """
    Poisson probability

    Args:
        x (int)     : measured number of occurences
        lmbda (int) : expected number of occurences

    Returns:
        np.ndarray
    """
    x = np.asarray(x, dtype=np.int32)
    pois = st.poisson(lmbda)
    return pois.pmf(x)
    #return np.power(lmbda, k) * np.exp(-1 * lmbda) / factorial(k)

################################################

def pandel_factory(c_ice):
    """
    Create a pandel function with the defined parameters. The pandel function
    is very specific, and a parametrisation for the delaytime distribution of 
    photons from a source s measured at a reciever r after traversing a certain
    large (compared to the size of source or reciever) distance in a homogenous
    scatterint medium such as ice or water.
    The version here has a number of fixed parameters optimized for IceCube.
    This function will generate a Pandel function with a single free parameter, 
    which is the distance between source and reciever.

    Args:
        c_ice (float): group velocity in ice in m/ns

    Returns:
        callable (float, float) -> float
    """
    tau = 450.
    el = 47.
    ella = 98.

    if c_ice > 1:
        Logger.warning("Wrong unit for c_ice... multiplying with 1e-9")
        c_ice *= 1e-9

    from math import gamma
    gamma = np.vectorize(gamma)
    a = lambda tau, ella: 1 / tau + c_ice / ella
    b = lambda r, el: r / el

    def pandel(x, distance):
        """

        Args:
            x : time where the pandel is evalueated (in ns)
        """
        val = (a(tau, ella) * ((a(tau, ella) * x) ** (b(distance, el) - 1))\
               * np.exp(-1 * a(tau, ella) * x)) / gamma(b(distance, el))
        return val

    return pandel


################################################

def gauss(x, mu, sigma):
    """
    Returns a normed gaussian.

    Args:
        x    (np.ndarray): x values
        mu        (float): Gauss mu
        sigma     (float): Gauss sigma
        n:

    Returns:

    """

    def _gaussnorm(sigma):
        return 1 / (sigma * np.sqrt(2 * np.pi))

    return _gaussnorm(sigma) * np.exp(-np.power((x - ( mu)), 2) / (2 *  (sigma ** 2)))



################################################

def n_gauss(x, mu, sigma, n):
    """
    Returns a normed gaussian in the case of n ==1. If n > 1, The gaussian
    mean is shifted by n and its width is enlarged by the factor of n.
    The envelope of a sequence of these gaussians will be an expoenential.

    Args:
        x (np.ndarray): x values
        mu (float): Gauss mu
        sigma (float): Gauss sigma
        n (int): > 0, linear coefficient

    Returns:

    """
    assert n > 0, "Can not compute gauss with n <= 0"

    def _gaussnorm(sigma, n):
        return 1 / (sigma * np.sqrt(2 * n * np.pi))

    return _gaussnorm(sigma, n) * np.exp(-np.power((x - (n * mu)), 2) / (2 * n * (sigma ** 2)))

####################################################

def calculate_sigma_from_amp(amp):
    """
    Get the sigma for the gauss from its peak value.
    Gauss is normed

    Args:
        amp (float):

    Returns:
        float
    """
    return 1/(np.sqrt(2*np.pi)*amp)

#####################################3

def exponential(x, lmbda):
    """
    An exponential model, e.g. for a decay with coefficent lmbda.

    Args:
        x (float): input
        lmbda (float): The exponent of the exponential

    Returns:
        np.ndarray
    """
    return np.exp(-lmbda * x)

#####################################################

def calculate_chi_square(data, model_data):
    """
    Very simple estimator for goodness-of-fit. Use with care.
    Non normalized bin counts are required.

    Args:
        data (np.ndarray): observed data (bincounts)
        model_data (np.ndarray): model predictions for each bin

    Returns:
        np.ndarray
    """
    #FIXME there might be a reason I had chosen 
    #      for going with my own implenetation 
    #      initially
    mask = np.logical_and(np.isfinite(data), np.isfinite(model_data))
    chi2 =  (data[mask] - model_data[mask])**2/data[mask]
    chi2 =  (chi2[np.isfinite(chi2)].sum())
    #chi2 = st.chisquare(scale, model_data)[0]
    Logger.warn(f'Calculated chi2 of {chi2}')
    return chi2
    #return chi[np.isfinite(chi)].sum()

#################################################

def calculate_reduced_chi_square(data, model_data, sigma):
    """
    Very simple estimator for goodness-of-fit. Use with care.
    
    Args:
        data (np.ndarray)       : observed data 
        model_data (np.ndarray) : model predictions 
        sigma (np.ndarray)      : associated errors
    Returns:
    """
    #FIXME there might be a reason I had chosen 
    #      for going with my own implenetation 
    #      initially
    mask = np.logical_and(np.isfinite(data), np.isfinite(model_data), sigma > 0)
    chi2 =  (data[mask] - model_data[mask])**2/sigma[mask]
    chi2 =  (chi2[np.isfinite(chi2)].sum())
    #chi2 = st.chisquare(scale, model_data)[0]
    Logger.debug(f'Calculated chi2 of {chi2}')
    return chi2

#################################################

def williams_correction():
    """
    The so-called Williams correction can help to correct
    a chi2 value in case of bins with low statistics (< 5 entries)
    """
    raise NotImplementedError

#################################################

def fwhm_gauss(x, mu, fwhm, amp):
    """
    A gaussian typically used for energy spectra fits of radiotion, where
    resolutions/linewidths are typically given in full widht half maximum (fwhm)

    Args:
        x (float)    : input
        mu (float)   : peak position
        fwhm (float) : full width half maximum
        amp (float)  : amplitude
    Return:
        float        : function value
    """
    return amp*np.exp((-4*np.log(2)*((x-mu)**2))/(fwhm**2))

################################################



"""
Define some
"""
from __future__ import print_function

from builtins import range
from builtins import object

from copy import deepcopy as copy

import numpy as n
import numpy as np
import dashi as d
import pylab as p

import matplotlib.ticker

from hepbasestack.colors import get_color_palette
#from .colors import get_color_palette
from .canvases import YStackedCanvas
from ..utils import Logger
from ..utils import flatten
from .. import fitting as fit

d.visual()

###############################################

def create_arrow(ax, x_0, y_0, dx, dy, length,\
                 width = .1, shape="right",\
                 fc="k", ec="k",\
                 alpha=1., log=False):
    """
    Create an arrow object for plots. This is typically a large
    arrow, which can used to indicate a region in the plot which
    is excluded by a cut.

    Args:
        ax (matplotlib.axes._subplots.AxesSubplot): The axes where the arrow
                                                    will be attached to
        x_0 (float): x-origin of the arrow
        y_0 (float): y-origin of the arrow
        dx (float): x length of the arrow
        dy (float): y length of the arrow
        length (float): additional scaling parameter to scale
                        the length of the arrow

    Keyword Args:
        width (float): thickness of arrow
        shape (str): either "full", "left" or "right"
        fc (str): facecolor
        ec (str): edgecolor
        alpha (float): 0 -1 alpha value of the arrow
        log (bool): I for logscale, the proportions of the arrow will be adjusted accorginly.

    Returns:
        matplotlib.axes._subplots.AxesSubplot
    """
    head_starts_at_zero = False
    head_width = width*5
    head_length = length*0.1
    if log: # slightly bigger arrow
        head_width = width*6
        head_length = length*0.2

    arrow_params={'length_includes_head':False,\
                  'shape':shape,\
                  'head_starts_at_zero':head_starts_at_zero}
    arr = ax.arrow(x_0, y_0, dx*length,\
                   dy*length, fc=fc, ec=ec,\
                   alpha=alpha, width=width,\
                   head_width=head_width,\
                   head_length=head_length,\
                   **arrow_params)
    return ax

###############################################

def meshgrid(xs, ys):
    """
    Create x and y data for matplotlib pcolormesh and
    similar plotting functions.

    Args:
        xs (np.ndarray): 1d x bins
        ys (np.ndarray): 2d y bins

    Returns:
        tuple (np.ndarray, np.ndarray, np.ndarray): 2d X and 2d Y matrices as well as a placeholder for the Z array

    """
    xlen = len(xs)
    ylen = len(ys)
    allx, ally = [], []
    
    # prepare xs
    for __ in range(ylen):
        allx.append(xs)

    allx = np.array(allx)
    allx = allx.T

    # prepare ys
    for __ in range(xlen):
        ally.append(ys)

    ally = np.array(ally)

    zs = np.zeros([xlen, ylen])
    return allx, ally, zs

###############################################

def line_plot(quantities,
              bins=None,
              xlabel='',
              add_ratio=None,
              ratiolabel='',
              colors=None,
              figure_factory=None):
    """
    Args:
        quantities:

    Keyword Args:
        bins:
        xlabel:
        add_ratio (tuple): (["data1"],["data2"])
        ratiolabel (str):
        colors:
        figure_factory (callable): Factory function returning matplotolib.Figure

    Returns:

    """
    # FIXME XXX under development
    raise NotImplementedError

    if add_ratio is not None:
        canvas = YStackedCanvas(subplot_yheights=(0.2, 0.7),
                                space_between_plots=0.0)
        ax0 = canvas.select_axes(-1)
        data = np.array(quantities[add_ratio[0]]) / np.array(quantities[add_ratio[1]])
        data = data * 100
        thebins = np.array(bins[add_ratio[0]])
        bin_size = abs(thebins[1] - thebins[0]) / 2 * np.ones(len(thebins))
        thebins = thebins + bin_size
        ax0.plot(thebins, data, color='gray')
        ax0.scatter(thebins, data, marker='o', s=50, color='gray')
        ax0.grid(1)
        ax0.set_ylabel(ratiolabel)
        ax0.spines['top'].set_visible(False)
        ax0.spines['bottom'].set_visible(False)
        ax0.spines['right'].set_visible(False)
        ax = canvas.select_axes(0)
        lgax = ax0
    else:
        if figure_factory is None:
            fig = p.figure()
        else:
            fig = figure_factory()
        ax = fig.gca()
        lgax = ax

    for reconame in quantities:
        thebins = np.array(bins[reconame])
        bin_size = abs(thebins[1] - thebins[0]) / 2 * np.ones(len(thebins))
        thebins = thebins + bin_size
        label = reconame.replace('-', '')
        if colors is not None:
            ax.plot(thebins, quantities[reconame], label=label, color=colors[reconame])
            ax.scatter(thebins, quantities[reconame], marker='o', s=50, c=colors[reconame])
        else:
            ax.plot(thebins, quantities[reconame], label=label)
            ax.scatter(thebins, quantities[reconame], marker='o', s=50)

    legend_kwargs = {'bbox_to_anchor': [0.0, 1.0, 1.0, 0.102],'loc': 3,
       'frameon': False,
       'ncol': 2,
       'borderaxespad': 0,
       'mode': 'expand',
       'handlelength': 2,
       'numpoints': 1
       }
    if len(quantities.keys()) == 3:
        legend_kwargs['ncol'] = 3
    ax.grid(1)
    ax.set_ylabel('$\\cos(\\Psi)$')
    ax.set_xlabel(xlabel)
    sb.despine()
    if add_ratio:
        ax.spines['top'].set_visible(False)
        ax0.spines['bottom'].set_visible(False)
        ax0.spines['left'].set_visible(False)
        legend_kwargs['bbox_to_anchor'] = [0, 1.3, 1.0, 0.102]
        ax.legend(**legend_kwargs)
        p.subplots_adjust(hspace=0.2)
        return canvas.figure
    ax.legend(**legend_kwargs)
    return fig

###############################################

def gaussian_model_fit(data,
                       startparams=(0,0.2),
                       fitrange=((None,None), (None,None)),
                       fig=None,
                       norm=True,
                       bins=80,
                       xlabel='$\\theta_{{rec}} - \\theta_{{true}}$'):
    """
    A plot with a gaussian fitted to data. A histogram of the data will be created and a gaussian
    will be fitted, with 68 and 95 percentiles indicated in the plot.

    Args:
        data (array-like)        : input data with a (preferably) gaussian distribution

    Keyword Args:
        startparams (tuple)      : a set of startparams of the gaussian fit. If only
                                   mu/sigma are given, then the plot will be normalized
        fig (matplotlib.Figure)  : pre-created figure to draw the plot in 
        bins (array-like or int) : bins for the underliying histogram
        fitrange (tuple(min, max): min-max range for the gaussian fit
        xlabel (str)             : label for the x-axes
    """
    if len(startparams) == 3:
        tofit = lambda x,mean,sigma,amp : amp*fit.gauss(x,mean,sigma)
    else:
        tofit = fit.gauss
    mod = fit.Model(tofit, startparams=startparams)
    mod.add_data(data, create_distribution=True, bins=bins, normalize=norm)
    limits = []
    notempty = False
    for k in fitrange:
        thislimit = []
        for j in k:
            if j is None:
                continue
            else:
                notempty = True
                thislimit.append(j)
        limits.append(tuple(thislimit))
    limits = tuple(limits)
    print (limits)
    if notempty:
        mod.fit_to_data(limits=limits)

    else:
        mod.fit_to_data()


    thecolors = get_color_palette()
    fig = mod.plot_result(log=False, xlabel=xlabel, add_parameter_text=(
     ('$\\mu$& {:4.2e}\\\\', 0), ('$\\sigma$& {:4.2e}\\\\', 1)), datacolor=thecolors[3], modelcolor=thecolors[3], histostyle='line', model_alpha=0.7, fig=fig)
    ax = fig.gca()
    ax.grid(1)
    ax.set_ylim(ymax=1.1 * max(mod.data))
    upper68 = mod.distribution.stats.mean + mod.distribution.stats.std
    lower68 = mod.distribution.stats.mean - mod.distribution.stats.std
    lower95 = mod.distribution.stats.mean - 2 * mod.distribution.stats.std
    upper95 = mod.distribution.stats.mean + 2 * mod.distribution.stats.std
    ax.axvspan(lower68, upper68, facecolor=thecolors[8], alpha=0.7, ec='none')
    ax.axvspan(lower95, upper95, facecolor=thecolors[8], alpha=0.3, ec='none')
    ax.text(lower68 * 0.9, max(mod.data) * 0.98, '68\\%', color=thecolors[3], fontsize=20)
    ax.text(lower95 * 0.9, max(mod.data) * 0.85, '95\\%', color=thecolors[3], fontsize=20)
    return (mod, fig)


###############################################

def gaussian_fwhm_fit(data,
                      startparams=(0,0.2,1),\
                      fitrange=((None,None), (None,None), (None, None)),\
                      fig=None,\
                      bins=80,\
                      xlabel='$\\theta_{{rec}} - \\theta_{{true}}$'):
    """
    A plot with a gaussian fitted to data. A histogram of the data will be created and a gaussian
    will be fitted, with 68 and 95 percentiles indicated in the plot. The gaussian will be in a form
    so that the fwhm can be read directly from it. The "width" parameter of the gaussian is NOT the
    standard deviation, but FWHM!

    Args:
        data (array-like)        : input data with a (preferably) gaussian distribution

    Keyword Args:
        startparams (tuple)      : a set of startparams of the gaussian fit. It is a 3
                                   parameter fit with mu, fwhm and amplitude
        fitrange (tuple)         : if desired, the fit can be restrained. One tuple of (min, max) per
                                   parameter
        fig (matplotlib.Figure)  : pre-created figure to draw the plot in
        bins (array-like or int) : bins for the underliying histogram
        xlabel (str)             : label for the x-axes
    """
    mod = fit.Model(fit.fwhm_gauss, startparams=startparams)
    mod.add_data(data, create_distribution=True, bins=bins, normalize=False)
    limits = []
    notempty = False
    for k in fitrange:
        thislimit = []
        for j in k:
            if j is None:
                continue
            else:
                notempty = True
                thislimit.append(j)
        limits.append(tuple(thislimit))
    limits = tuple(limits)
    print (limits)
    if notempty:
        mod.fit_to_data(limits=limits)

    else:
        mod.fit_to_data()

    thecolors = get_color_palette()
    fig = mod.plot_result(log=False, xlabel=xlabel, add_parameter_text=(
        ('$\\mu$& {:4.2e}\\\\', 0), ('FWHM& {:4.2e}\\\\', 1), ('AMP& {:4.2e}\\\\',2)), datacolor=thecolors[3], modelcolor=thecolors[3], histostyle='line', model_alpha=0.7, fig=fig)
    ax = fig.gca()
    ax.grid(1)
    ax.set_ylim(ymax=1.1 * max(mod.data))
    upper68 = mod.distribution.stats.mean + mod.distribution.stats.std
    lower68 = mod.distribution.stats.mean - mod.distribution.stats.std
    lower95 = mod.distribution.stats.mean - 2 * mod.distribution.stats.std
    upper95 = mod.distribution.stats.mean + 2 * mod.distribution.stats.std
    ax.axvspan(lower68, upper68, facecolor=thecolors[8], alpha=0.7, ec='none')
    ax.axvspan(lower95, upper95, facecolor=thecolors[8], alpha=0.3, ec='none')
    ax.text(lower68 * 0.9, max(mod.data) * 0.98, '68\\%', color=thecolors[3], fontsize=20)
    ax.text(lower95 * 0.9, max(mod.data) * 0.85, '95\\%', color=thecolors[3], fontsize=20)
    return (mod, fig)


###############################################

class VariableDistributionPlot(object):
    """
    A plot which shows the distribution of a certain variable.
    Cuts can be indicated with lines and arrows. This class defines
    (and somehow enforces) a certain style.
    """

    def __init__(self, cuts=None,\
                 color_palette="dark",\
                 bins=None,\
                 xlabel=None):
        """
        Keyword Args:
            bins               (array-like): desired binning, if None use default
            cuts (HErmes.selection.cut.Cut): conditions applied to the dataset.
                                             the cut will not be performed, but visualized
                                             instead with a line and a arrow indicatng the
                                             cutted region
            color_palette             (str): use this palette for the plotting
            xlabel                    (str): descriptive string for x-label
        """
        self.histograms = {}
        self.histratios = {}
        self.cumuls = {}
        self.plotratio = False
        self.plotcumul = False
        self.canvas = None
        if (xlabel is None):
            self.label = ''
        else:
            self.label = xlabel
        self.name = ''
        self.bins = bins
        if cuts is None:
            cuts = []
        self.cuts = cuts
        if isinstance(color_palette, str):
            self.color_palette = get_color_palette(color_palette)
        else:
            self.color_palette = color_palette
        self.plot_options = dict()

    def add_cuts(self, cut):
        """
        Add a cut to the the plot which can be indicated by an arrow

        Args:
            cuts (HErmes.selection.cuts.Cut):

        Returns:
            None
        """
        self.cuts.append(cut)

    def add_data(self, variable_data,\
                 name, bins=None,\
                 weights=None, label=''):
        """
        Histogram the added data and store internally
        
        Args:
            name (string): the name of a category
            variable_data (array): the actual data
        
        Keyword Args:
            bins (array): histogram binning
            weights (array): weights for the histogram
            label (str): A label for the data when plotted

        """
        if bins is None:
            bins = self.bins
        if weights is None:
            self.histograms[name] = d.factory.hist1d(variable_data, bins)
        else:
            Logger.debug(f"Found {len(weights)} weights and {len(variable_data)} data points")
            assert len(weights) == len(variable_data),\
                 f"Mismatch between len(weights) {len(weights)} and len(variable_data) {len(variable_data)}"
            self.histograms[name] = d.factory.hist1d(variable_data, bins, weights=weights)
        self.label = label
        self.name = name

    def add_variable(self, category,
                     variable_name,
                     external_weights=None,
                     transform=None):
        """
        Convenience interface if data is sorted in categories already

        Args:
            category (HErmese.variables.category.Category): Get variable from this category
            variable_name (string): The name of the variable

        Keyword Args:
            external_weights (np.ndarray): Supply an array for weighting. This will OVERIDE ANY INTERNAL WEIGHTING MECHANISM and use the supplied weights.
            transform (callable): Apply transformation todata

        """
        if category.plot_options:
            self.plot_options[category.name] = copy(category.plot_options)
        if self.bins is None:
            self.bins = category.vardict[variable_name].bins
        self.name = variable_name
        if external_weights is None:
            Logger.warning("Internal weighting mechanism is broken at the moment, FIXME!")
            #weights = category.weights
            #if len(weights) == 0:
            #    weights = None
            weights = None
        else:
            weights = external_weights
        if transform is None: transform = lambda x : x
        data = category.get(variable_name)
        #print (variable_name)
        #print (data)
        #print (data[0])
        # FIXME: check this
        # check pandas series and 
        # numpy array difference
        # FIMXME: values was as_matrix before - adapt changes in requirements.txt
        try:
            data = data.values
        except:
            pass
        # hack for applying the weights
        if hasattr(data[0],"__iter__"):
            if weights is not None:
                Logger.warning("Multi array data for {} de tected. Trying to apply weights".format(variable_name))
                tmpweights = np.array([weights[i]*np.ones(len(data[i])) for i in range(len(data))])
                Logger.warning("Weights broken, assuming flatten as transformation")
                weights = flatten(tmpweights)

        self.add_data(transform(data),\
                      category.name,\
                      self.bins, weights=weights,\
                      label=category.vardict[variable_name].label)
        
    def add_cumul(self, name):
        """
        Add a cumulative distribution to the plot

        Args:
            name (str): the name of the category
        """
        assert name in self.histograms, "Need to add data first"
        self.cumuls[name] = self.histograms[name].normalized()

    def indicate_cut(self, ax, arrow=True):
        """
        If cuts are given, indicate them by lines

        Args:
            ax (pylab.axes): axes to draw on

        """
        vmin, vmax = ax.get_ylim()
        hmin, hmax = ax.get_xlim()
        for cut in self.cuts:
            for name, (operator, value) in cut:

                # there might be more than one
                # cut which should be
                # drawn on this plot
                # so go through ALL of them.
                if name != self.name:
                    continue

                Logger.debug('Found cut! {0} on {1}'.format(name,value))
                width = vmax/50.

                # create a line a cut position
                ax.vlines(value, ymin=vmin, ymax=vmax, linestyle=':')
                length = (hmax - hmin)*0.1

                # mark part of the plot as "forbidden"
                # and create arrow if desired
                if operator in ('>','>='):
                    shape = 'right'
                    ax.axvspan(value, hmax,\
                               facecolor=self.color_palette["prohibited"],\
                               alpha=0.5)
                else:
                    shape = 'left'
                    ax.axvspan(hmin, value,\
                               facecolor=self.color_palette["prohibited"],\
                               alpha=0.5)

                if arrow:
                    ax = create_arrow(ax, value, vmax*0.1,\
                                      -1., 0, length, width=width,\
                                      shape=shape, log=True)

    def add_ratio(self, nominator, denominator,\
                  total_ratio=None, total_ratio_errors=None, \
                  log=False, label="data/$\Sigma$ bg"):
        """
        Add a ratio plot to the canvas

        Args:
            nominator (list or str): name(s) of the categorie(s) which
                                     will be the nominator in the ratio
            denominator (list or str): name(s) of the categorie(s) which
                                     will be the nominator in the ratio

        Keyword Args:
            total_ratio (bool): Indicate the total ratio with a line in the plot
            total_ratio_errors (bool): Draw error region around total ratio
            log (bool): draw ratio plot in log-scale
            label (str): y-label for the ratio plot

        """
        if not isinstance(nominator, list):
            nominator = [nominator]
        if not isinstance(denominator, list):
            denominator = [denominator]

        name = "".join(nominator) + "_" + "".join(denominator)
        first_nominator = nominator.pop()
        nominator_hist = self.histograms[first_nominator]
        nominator_ws   = self.histograms[first_nominator].stats.weightsum
        for name in nominator:
            nominator_hist += self.histograms[name]
            nominator_ws   += self.histograms[name].stats.weightsum

        first_denominator = denominator.pop()
        denominator_hist = self.histograms[first_denominator]
        denominator_ws = self.histograms[first_denominator].stats.weightsum
        for name in denominator:
            denominator_hist += self.histograms[name]
            denominator_ws   += self.histograms[name].stats.weightsum
    
        nominator_hist.normalized()
        denominator_hist.normalized()
        ratio = d.histfuncs.histratio(nominator_hist, denominator_hist,\
                                      log=False, ylabel=label)
        if total_ratio is None:
            total_ratio = nominator_ws/denominator_ws
            Logger.info("Calculated scalar ratio of {:4.2f} from histos".format(total_ratio))

        #ratio.y[ratio.y > 0] = ratio.y[ratio.y > 0] + total_ratio -1
        self.histratios[name] = (ratio, total_ratio, total_ratio_errors,\
                                 label)
        return name

    def _draw_distribution(self, ax, name,
                           log=True,\
                           cumulative=False,
                           normalized=False, 
                           ylabel="rate/bin [1/s]"):
        """
        Paint the histograms!
        
        Args:    
    
        Keyword Args:
            normalized (bool): draw by number of events normalized distribution
            log        (bool): Use a logarithmic scale for the y-axis                  
            cumulative (bool): Show a cumulative distribution 
            normalized (bool): normalize by number of entries
            ylable      (str): shown label on y-axis
    
        """
        try:
            cfg = copy(self.plot_options[name])
        except KeyError:
            Logger.warning("No plot configuration available for {}".format(name))
            cfg = {"histotype": "line",
                   "label": name,
                   "linestyle" : {"color": "k",
                                  "linewidth": 3
                                  }
                   }
        log = False
        if "linestyle" in cfg: 
            color = cfg["linestyle"].pop('color')
            if isinstance(color, int):
                color = self.color_palette[color]
        if 'scatterstyle' in cfg:
            scattercolor = cfg["scatterstyle"].pop('color')

            if isinstance(scattercolor,int):
                scattercolor = self.color_palette[scattercolor]

        if cumulative:
            histograms = self.cumuls
            log = False
        else:
            histograms = self.histograms

        if normalized and not cumulative:
            histograms[name] = histograms[name].normalized()

        if cfg['histotype'] == 'scatter':
            histograms[name].scatter(log=log,cumulative=cumulative,\
                                     label=cfg["label"],\
                                     color=scattercolor, **cfg["scatterstyle"])
        elif cfg['histotype'] == "line":
            # apply th alpha only to the "fill" setting
            linecfg = copy(cfg["linestyle"])
            if "alpha" in linecfg:
                linecfg.pop("alpha")
                linecfg["filled"] = False
            if "filled" in cfg["linestyle"]:
                histograms[name].line(log=log, cumulative=cumulative,\
                                  label=cfg["label"], color=color,\
                                  **linecfg)

                histograms[name].line(log=log, cumulative=cumulative,\
                                  label=None, color=color,\
                                  **cfg["linestyle"])
            else:
                histograms[name].line(log=log, cumulative=cumulative,\
                                  label=cfg["label"], color=color,\
                                  **cfg["linestyle"])
        elif cfg['histotype'] == "overlay":
            histograms[name].line(log=log, cumulative=cumulative,\
                                  label=cfg["label"], color=color,\
                                  **cfg["linestyle"])
            histograms[name].scatter(log=log, cumulative=cumulative,\
                                     color=scattercolor,\
                                     **cfg["scatterstyle"])
        if cumulative:
            ax.set_ylabel('fraction')
        else:
            ax.set_ylabel(ylabel)

    def _draw_histratio(self, name, axes, ylim=(0.1,2.5)):
        """
        Plot one of the ratios
        
        Returns:
            tuple (float,float) : the min and max of the ratio
        """
        ratio,total_ratio,total_ratio_errors,label = self.histratios[name]
        ratio.scatter(c="k", marker="o", markersize=3)
        axes.hlines(total_ratio,axes.get_xlim()[0],axes.get_xlim()[1],linestyle="--")
        if total_ratio_errors is not None:
            axes.hlines(total_ratio + total_ratio_errors,axes.get_xlim()[0],axes.get_xlim()[1],linestyle=":")
            axes.hlines(total_ratio - total_ratio_errors,axes.get_xlim()[0],axes.get_xlim()[1],linestyle=":")
            xs = n.linspace(axes.get_xlim()[0],axes.get_xlim()[1],200)
            axes.fill_between(xs,total_ratio - total_ratio_errors,\
                              total_ratio + total_ratio_errors,\
                              facecolor="grey", alpha=0.3)
        #axes.set_ylim(ylim)
        axes.set_ylabel(label)
        axes.grid(1)
        if not ratio.y is None:
            if not ratio.yerr is None:
                ymin = min(ratio.y - ratio.yerr)
                ymax = max(ratio.y + ratio.yerr)
                ymin -= (0.1*ymin)
                ymax += (0.1*ymax)
                return ymin, ymax
            else:
                ymin = min(ratio.y)
                ymax = max(ration.y)
                ymin -= (0.1*ymin)
                ymax += (0.1*ymax)
                return ymin, ymax
        else:
            return 0,0

    def _locate_axes(self, combined_cumul, combined_ratio, combined_distro):
        axes_locator = []
        if self.cumuls:
            if combined_cumul:
                axes_locator.append((0, "c"))
            else:
                axes_locator += [(x,"c") for x in range(len(self.cumuls))]

        if self.histratios:
            if combined_ratio:
                if axes_locator:
                    axes_locator.append((axes_locator[-1] + 1,"r"))
                else:
                    axes_locator.append((0,"r"))
            else:
                if axes_locator:
                    axes_locator +=[(x+ len(axes_locator),"r") for x  in range(len(self.histratios))]

        if self.histograms:
            if combined_distro:
                if axes_locator:
                    axes_locator.append((axes_locator[-1] + 1,"h"))
                else:
                    axes_locator.append((0,"h"))
            else:
                if axes_locator:
                    axes_locator +=[(x+ len(axes_locator),"h") for x  in range(len(self.histograms))]
        return axes_locator

    def plot(self,
             axes_locator=((0, "c",.2), (1, "r",.2), (2, "h",.5)),\
             combined_distro=True,\
             combined_ratio=True,\
             combined_cumul=True,
             normalized=True,
             style="classic",\
             log=True,
             legendwidth = 1.5,
             ylabel="rate/bin [1/s]",
             figure_factory=None,
             zoomin=False,
             adjust_ticks=lambda x : x):
        """
        Create the plot

        Keyword Args:
            axes_locator (tuple): A specialized tuple defining where the axes should be located in the plot
                                  tuple has the following form: 
                                  ( (PLOTA), (PLOTB), ...) where PLOTA is a tuple itself of the form (int, str, int)
                                  describing (plotnumber, plottype, height of the axes in the figure)
                                  plottype can be either: "c" - cumulative
                                                          "r" - ratio
                                                          "h" - histogram
            combined_distro:
            combined_ratio:
            combined_cumul:
            log (bool):
            style (str): Apply a simple style to the plot. Options are "modern" or "classic"
            normalized (bool):
            figure_factor (fcn): Must return a matplotlib figure, use for custom formatting
            zoomin (bool): If True, select the yrange in a way that the interesting part of the 
                           histogram is shown. Caution is needed, since this might lead to an
                           overinterpretation of fluctuations.
            adjust_ticks (fcn): A function, applied on a matplotlib axes
                                which will set the proper axis ticks


        Returns:

        """

        Logger.info(f"Found {len(self.histograms)} distributions")
        Logger.info(f"Found {len(self.histratios)} ratios")
        Logger.info(f"Found {len(self.cumuls)} cumulative distributions")
        if not axes_locator:
            axes_locator = self._locate_axes(combined_cumul,\
                                             combined_ratio,\
                                             combined_distro)

        # calculate the amount of needed axes
        # assert len(axes_locator) == len(heights), "Need to specify exactly as many heights as plots you want to have"

        heights = [k[2] for k in axes_locator]
        self.canvas = YStackedCanvas(subplot_yheights=heights,\
                                     figure_factory=figure_factory)
        
        cu_axes = [x for x in axes_locator if x[1] == "c"]
        h_axes  = [x for x in axes_locator if x[1] == "h"]
        r_axes  = [x for x in axes_locator if x[1] == "r"]

        maxheights = []
        minheights = []
        for ax in cu_axes:
            cur_ax = self.canvas.select_axes(ax[0])
            if combined_cumul:
                for k in list(self.cumuls.keys()):
                    self._draw_distribution(cur_ax, k, cumulative=True,log=log, ylabel=ylabel)
                break
            else:
                k = self.cumuls[list(self.cumuls.keys())[ax[0]]]
                self._draw_distribution(cur_ax,cumulative=True,log=log, ylabel=ylabel)


        for ax in r_axes:
            cur_ax = self.canvas.select_axes(ax[0])
            if combined_ratio:
                for k in list(self.histratios.keys()):
                    ymin, ymax = self._draw_histratio(k,cur_ax)
                    ymin -= (ymin*0.1)
                    ymax += (ymax*0.1)
                    cur_ax.set_ylim(ymin,ymax)

                    # FIXME: good tick spacing
                    #major_tick_space = 1
                    #minor_tick_space = 0.1
                    ## in case there are too many ticks
                    #nticks = float(ymax - ymin)/major_tick_space
                    #while nticks > 4:
                    #    major_tick_space += 1 

                    ##if ymax - ymin < 1:
                    ##    major_tick_space =  
                    #cur_ax.xaxis.set_major_locator(matplotlib.ticker.MultipleLocator(minor_tick_space))
                break
            else:
                k = self.histratios[list(self.histratios.keys())[ax[0]]]
                ymin, ymax = self._draw_histratio(k,cur_ax)    
                ymin -= (ymin*0.1)
                ymax += (ymax*0.1)
                cur_ax.set_ylim(ymin,ymax)

        for ax in h_axes:
            cur_ax = self.canvas.select_axes(ax[0])
            if combined_distro:
                for k in list(self.histograms.keys()):
                    Logger.debug("drawing..{}".format(k))
                    self._draw_distribution(cur_ax,k,log=log, normalized=normalized, ylabel=ylabel)
                break
            else:
                k = self.histograms[list(self.histograms.keys())[ax[0]]]
                ymax, ymin = self._draw_distribution(cur_ax,k,log=log, normalized=normalized, ylabel=ylabel)
            cur_ax.set_ylim(ymin=ymin - 0.1*ymin,ymax=1.1*ymax)
            cur_ax.grid(True)
        lgax = self.canvas.select_axes(-1) # most upper one
        ncol = 2 if len(self.histograms) <= 4 else 3 
        if style == "classic":
            # draw the legend in the box above the plot
            legend_kwargs = {"bbox_to_anchor": [0., 1.0, 1., .102],
                             "loc": 3,
                             "frameon": True,
                             "ncol": ncol,
                             "framealpha": 1.,
                             "borderaxespad": 0,
                             "mode": "expand",
                             "handlelength": 2,
                             "numpoints": 1}
            lg = lgax.legend(**legend_kwargs)
            if lg is not None:
                lg.get_frame().set_linewidth(legendwidth)
                lg.get_frame().set_edgecolor("k")
            else:
                Logger.warning("Can not set legendwidth!")       
        if style == "modern":
            # be more casual
            lgax.legend() 

        # plot the cuts
        if self.cuts:
            for ax in h_axes:
                cur_ax = self.canvas.select_axes(ax[0])
                self.indicate_cut(cur_ax, arrow=True)
            for ax in r_axes + cu_axes:
                cur_ax = self.canvas.select_axes(ax[0])
                self.indicate_cut(cur_ax, arrow=False)
        # cleanup
        leftplotedge, rightplotedge, minplotrange, maxplotrange = self.optimal_plotrange_histo(self.histograms.values())
        if minplotrange == maxplotrange:
            Logger.debug("Detected histogram with most likely a single bin!")
            Logger.debug("Adjusting plotrange")
            
        else:
            if zoomin: 
                figure_span = maxplotrange - minplotrange
                minplotrange -= (figure_span*0.1)
                maxplotrange += (figure_span*0.1) 
            else: # start at zero and show the boring part
                minplotrange = 0
                maxplotrange += (maxplotrange*0.1)
        if log:
            maxplotrange = 10**(np.log10(maxplotrange) + 1)

            if maxplotrange < 1: 
                minplotrange -= (minplotrange*0.01)
            else:
                minplotrange = 0 # will be switched to symlog by default

        for ax in h_axes:
            self.canvas.select_axes(ax[0]).set_ylim(ymin=minplotrange, ymax=maxplotrange)
            self.canvas.select_axes(ax[0]).set_xlim(xmin=leftplotedge, xmax=rightplotedge)
            if log:
                if maxplotrange < 1:
                    self.canvas.select_axes(ax[0]).set_yscale("log")
                else:
                    self.canvas.select_axes(ax[0]).set_yscale("symlog")

        for ax in cu_axes:
            self.canvas.select_axes(ax[0]).set_xlim(xmin=leftplotedge, xmax=rightplotedge)
        for ax in r_axes:
            self.canvas.select_axes(ax[0]).set_xlim(xmin=leftplotedge, xmax=rightplotedge)
        self.canvas.eliminate_lower_yticks()
        # set the label on the lowest axes
        self.canvas.axes[0].set_xlabel(self.label)
        minor_tick_space = self.canvas.axes[0].xaxis.get_ticklocs()
        minor_tick_space = (minor_tick_space[1] - minor_tick_space[0])/10.
        if minor_tick_space < 0.1:
            Logger.debug("Adjusting for small numbers in tick spacing, tickspace detectected {}".format(minor_tick_space))
            minor_tick_space = 0.1
        self.canvas.axes[0].xaxis.set_minor_locator(matplotlib.ticker.MultipleLocator(minor_tick_space))
        for x in self.canvas.axes[1:]:
            p.setp(x.get_xticklabels(), visible=False)
            p.setp(x.get_xticklines(), visible=False)
            x.xaxis.set_tick_params(which="both",\
                                    length=0,\
                                    width=0,\
                                    bottom=False,\
                                    labelbottom=False)
        for x in self.canvas.figure.axes:
            x.spines["right"].set_visible(True)
            adjust_ticks(x)

        for ax in h_axes:
            #self.canvas.select_axes(ax[0]).ticklabel_format(useOffset=False, style='plain', axis="y")
            self.canvas.select_axes(ax[0]).get_yaxis().get_offset_text().set_x(-0.1)

        if ((len(h_axes) == 1) and (style == "modern")):
            self.canvas.select_axes(-1).spines["top"].set_visible(False)
            self.canvas.select_axes(-1).spines["right"].set_visible(False)


    @staticmethod
    def optimal_plotrange_histo(histograms):
        """
        Get most suitable x and y limits for a bunc of histograms
        
        Args:
            histograms (list(d.factory.hist1d)): The histograms in question

        Returns:
            tuple (float, float, float, float): xmin, xmax, ymin, ymax

        """

        leftplotedge = n.inf
        rightplotedge = -n.inf
        minplotrange = n.inf
        maxplotrange = -n.inf
        for h in histograms:
            if not h.bincontent.any():
                continue
            if h.bincenters[h.bincontent > 0][0] < leftplotedge:
                leftplotedge = h.bincenters[h.bincontent > 0][0]
                leftplotedge -= h.binwidths[0]

            if h.bincenters[h.bincontent > 0][-1] > rightplotedge:
                rightplotedge = h.bincenters[h.bincontent > 0][-1]
                rightplotedge += h.binwidths[0] 

            if min(h.bincontent[h.bincontent > 0]) < minplotrange:
                minplotrange = min(h.bincontent[h.bincontent > 0])
                
            if max(h.bincontent[h.bincontent > 0]) > maxplotrange:
                maxplotrange = max(h.bincontent[h.bincontent > 0])

        Logger.info("Estimated plotrange of xmin {} , xmax {}, ymin {}, ymax {}".format(leftplotedge, rightplotedge, minplotrange, maxplotrange)) 
        return leftplotedge, rightplotedge, minplotrange, maxplotrange


    def add_legend(self, **kwargs):
        """
        Add a legend to the plot. If no kwargs are passed,
        use some reasonable default.
        
        Keyword Args:
             will be passed to pylab.legend
        """
        if not kwargs:
            kwargs = {"bbox_to_anchor": (0.,1.0, 1., .102),\
                      "loc" : 3, "ncol" :3,\
                      "mode": "expand",\
                      "framealpha": 1,\
                      "borderaxespad": 0.,\
                      "handlelength": 2,\
                      "numpoints": 1}
        self.canvas.global_legend(**kwargs)
 
# #######################################################
#
# def error_distribution_plot(h,
#                             xlabel = r"$\log(E_{rec}/E_{ref})$",
#                             name = "E",
#                             median = False):
#     """
#
#
#     Args:
#         h:
#         xlabel:
#         name:
#         median:
#
#     Returns:
#
#     """
#     par = HistoFitter(h, Gauss)
#     fig  = p.figure(figsize=(6,4),dpi=350)
#     ax   = fig.gca()
#     if not median: ax.plot(h.bincenters, Gauss(par,h.bincenters),color="k",lw=2)
#     h.line(filled=True,color="k",lw=2,fc="grey",alpha=.5)#hatch="//")
#     h.line(color="k",lw=2)
#     ax.grid(1)
#     ax.set_ylim(ymax=1.1*max(h.bincontent))
#     ax.set_xlim(xmin=h.bincenters[0],xmax=h.bincenters[-1])
#     ax.set_xlabel(xlabel)
#     ax.set_ylabel("Normalized bincount")
#     if median: ax.vlines(h.stats.median,0,1.1*max(h.bincontent),linestyles="dashed")
#     textstr ="Gaussian fit:\n"
#     textstr += "$\mu$ = " + "%4.3f" %par[1] + "\n" + "$\sigma$ = " + "%4.2f" %par[2]
#     if median:
#         textstr = "Median:\n %4.3f" %h.stats.median
#     CreateTextbox(ax,textstr,boxstyle="square",xcoord=.65,fontsize=16,alpha=.9)
#     #Thesisize(ax)
#     #print h.bincontent[h.bincenters > -.1][h.bincenters < .1].cumsum()[-1]
#
#     #ChisquareTest(h,Gauss(par,h.bincenters),xmin=-.001,xmax=.001)
#     #savename = Multisavefig(plotdir_stat,"parameter-reso-" + name,3,orientation="portrait",pad_inches=.3,bbox_inches="tight")[0]#pad_inches=.5,bbox_inche     s="tight")[0]
#     return fig
#
# ####################################################
#
# def HistoFitter(histo,func,startmean=0,startsigma=.2):
#
#     def error(p,x,y):
#         return n.sqrt((func(p,x) - y)**2)
#
#     #print histo.bincontent.std()
#     histo.stats.mean
#     p0 = [max(histo.bincontent),histo.stats.mean,histo.stats.var]
#     output = optimize.leastsq(error,p0,args=(histo.bincenters,histo.bincontent),full_output=1)
#     par = output[0]
#     covar = output[1]
#     rchisquare = scipy.stats.chisquare(1*histo.bincontent,f_exp=(1*func(par,histo.bincenters)))[0]/(1*(len(histo.bincenters) -len(par)))
#     #print par,covar
#     #print "chisquare/ndof",rchisquare
#     #print histo.bincontent[:10], func(par,histo.bincenters)[:10]
#     #print "ks2_samp", scipy.stats.ks_2samp(histo.bincontent,func(par,histo.bincenters))
#     return par
#
# #####################################################
#
# def create_textbox(ax, textstr, boxstyle="round",\
#                    facecolor="white", alpha=.7,\
#                    xcoord=0.05, ycoord=0.95, fontsize=14):
#     """
#     Create a textbox on a given axis
#
#     Args:
#         ax:
#         textstr:
#         boxstyle:
#         facecolor:
#         alpha:
#         xcoord:
#         ycoord:
#         fontsize:
#
#     Returns:
#         the given ax object
#     """
#     props = dict(boxstyle=boxstyle, facecolor=facecolor, alpha=alpha)
#     # place a text box in upper left in axes coords
#     ax.text(xcoord, ycoord, textstr,\
#             transform=ax.transAxes,\
#             fontsize=fontsize,\
#             verticalalignment='top', bbox=props)
#     return ax
#
# ######################################################
#
# def ChisquareTest(histo, fit, xmin=-.2, xmax=.2, ndof=3):
#     data = histo.bincontent[histo.bincenters > xmin][histo.bincenters < xmax]
#     fit  = fit[histo.bincenters > xmin][histo.bincenters < xmax]
#     #print data,fit
#     print (scipy.stats.chisquare(data,f_exp=fit)[0]/(len(fit) - ndof))
#     print (scipy.stats.ks_2samp(data,fit))
#     print (scipy.stats.anderson(data))





"""
A package for filtering datasets as common in high energy physics. Read data from hdf or
root files, classify the data in different categories and provide and easy interface
to easy access to the variables stored in the files.

The HErmes modules provides the following submodules:

- `selection` : Start from a .json configuration file to create a full fledged dataset which acts as a container for in different categories.

- `utils` : Aggregator for files and logging.

- `fitting` : Fit models to variable distributions with iminuit.

- `visual` : Data visualization.

- `icecube_goodies` : Weighting for icecube datasets.

- `analysis` : convenient functions for data analysis and working with distributions.

"""

__version__ = '0.1.1'
__all__ = ["fitting", "icecube_goodies", "utils",\
           "selection", "visual", "analysis"]

import os.path
import hepbasestack as hep

from . import utils
loglevel = hep.logger.LOGLEVEL

from . import visual

def set_loglevel(level):
    """
    Set the loglevel, 10 = debug, 20 = info, 30 = warn
    """
    hep.logger.LOGLEVEL = level
    return

# FIXME documentation
def _hook():
    pass



"""
Common calculations
"""

import numpy as np

def opening_angle(reco_zen, reco_azi, true_zen, true_azi):
    """
    Calculate the opening angle between two vectors, described 
    by azimuth and zenith in some coordinate system. 
    Can be useful for estimatiion of angular uncertainty 
    for some reconstruction.
    Zenith and Azimuth in radians.

    Args:
        reco_zen (float): zenith of vector A
        reco_azi (float): azimuth of vector A
        true_zen (float): zenith of vector B
        true_azi (float): azimuth of vector B

    Returns:
        float: Opening angle in degree
    """

    sinisy = lambda x,y : np.sin(x)*np.sin(y)
    cosisy = lambda x,y : np.cos(x)*np.cos(y)

    c = np.cos
    
    # dot product
    cospsi = ((c(reco_zen)*c(true_zen)) + (sinisy(reco_zen,true_zen)*cosisy(reco_azi,true_azi)) + (sinisy(reco_zen,true_zen)*sinisy(reco_azi,true_azi)))
    psi = np.arccos(cospsi)
    psi = np.rad2deg(psi)
    return psi







"""
Goodies for icecube
"""

import tables
import numpy as np
import os

GEOMETRY=os.path.join(os.path.split(__file__)[0],"geometry_ic86.h5")

class IceCubeGeometry(object):
    """
    Provide icecube geometry information
    """

    def __init__(self):
        self.geofile = None
        self.geo = None
        self.load_geo()

    def load_geo(self):
        """
        Load geometry information
        """

        #FIXME absolute path
        self.geofile = tables.open_file(GEOMETRY)
        self.geo = self.geofile.root.geometry.read()

    def coordinates(self, string, dom):
        """
        Calculate the xy position of a given string
        """
        #string -= 1

        assert string in range(1, 86), "String {} must be a number from 1 - 86".format(string)
        assert dom in range(1, 61), "Dom {} must be a number from 1 - 60".format(dom)

        mask = (self.geo['om'] == dom) & (self.geo['string'] == string)
        position = self.geo[mask]
        pos      = np.array([np.float(position['x']),np.float(position['y']),np.float(position['z'])])
        return pos

    def __del__(self):
        self.geofile.close()







"""
Provide a simple, easy to use model for fitting data and especially
distributions. The model is capable of having "components", which can
be defined and fitted individually.
"""


from functools import reduce

import numpy as np
import dashi as d
import pylab as p
import iminuit
import types
import functools
import inspect
from copy import deepcopy as copy

import scipy.optimize as optimize
import seaborn as sb

from . import functions as funcs
from .. import utils as u
Logger = u.Logger

default_color = "r"
try:
    default_color = sb.color_palette("dark")[2]
except Exception as e:
    Logger.warn("Can not use seaborn dark colorpalette for setting the default color! Exception thrown {}".\
                format(e))


def copy_func(f):
    """
    Based on http://stackoverflow.com/a/6528148/190597 (Glenn Maynard)

    Basically recreate the function f independently.

    Args:
        f (callable): the function f will be cloned
    """
    g = types.FunctionType(f.__code__, f.__globals__, name=f.__name__,
                           argdefs=f.__defaults__,
                           closure=f.__closure__)
    g = functools.update_wrapper(g, f)
    g.__kwdefaults__ = f.__kwdefaults__
    return g

def concat_functions(fncs):
    """
    Inspect functions and construct a new one which returns the added result.
    concat_functions(A(x, apars), B(x, bpars)) -> C(x, apars,bpars)
    C(x, apars, bpars) returns (A(x, apars) + B(x, bpars))

    Arguments:
        fncs (list):
            The callables to concat

    Returns:
        tuple (callable, list(pars))

    """

    datapar = inspect.getfullargspec(fncs[0]).args[0]
    pars = [inspect.getfullargspec(f).args[1:] for f in fncs]
    npars = [len(prs) for prs in pars]
    renamed_pars = [[k + str(i) for k in ipars] for i,ipars in enumerate(pars)]
    joint_pars = reduce( lambda x,y : x + y, renamed_pars)

    slices = []
    lastslice = 0
    for i, pr in enumerate(npars):
        thisslice = slice(lastslice, npars[i] + lastslice)
        slices.append(thisslice)
        lastslice += npars[i]
    globals().update({"concat_functions_fncs":fncs,\
                      "concat_functions_slices" : slices,\
                      "concat_functions_joint_pars" : joint_pars,\
                      "concat_functions_datapar" : datapar})
    TEMPLATE = """def jointfunc({}):
    data = concat_functions_fncs[0](*([{}] + ([{}][concat_functions_slices[0]])))
    for i,fn in enumerate(concat_functions_fncs[1:]):
        data += fn(*([{}] + ([{}][concat_functions_slices[1:][i]])))
    return data
""".format(datapar + "," + ",".join(joint_pars),datapar, ",".join(joint_pars),datapar, ",".join(joint_pars))
    exec (TEMPLATE, globals())
    return jointfunc, joint_pars

def construct_efunc(x, data, jointfunc, joint_pars):
    """
    Construct a least-squares error function. This function will then 
    be minimized, e.g. with the help of minuit.

    Args:
        x            (np.ndarray): The x-values the fit should be evaluated on
        data:        (np.ndarray): The y-values of the data we want to describe
        jointfunc:     (callable): The full data model with all components
        joint_pars:       (tuple): The model parameters   

    Returns:
        callable                 
    """
    # ensure that "x" is the first parameter, even if it 
    # has a different name
    datapar = inspect.getfullargspec(jointfunc).args[0]
    globals().update({"jointfunc" : jointfunc,\
                      "{}".format(datapar) : x,\
                      "data" : data})
    EFUNC = """def efunc({}):
    return ((abs(data - jointfunc({}))**2).sum())""".format(",".join(joint_pars), datapar + "," + ",".join(joint_pars))

    exec (EFUNC, globals())
    return efunc

def create_minuit_pardict(fn, startparams, errors, limits, errordef):
    """
    Construct a dictionary for minuit fitting. This dictionary contains information
    for the minuit fitter like startparams or limits.

    Args:
        fn        (callable): The function for which 
        startparams  (tuple): A list of startparameter. One each per parameter
        errors        (list): ?
        limits (list(tuple)): A list of (min, max) tuples for each parameter, can be None
        errordef     (float): The errordef should be 1 for a least square fit 
                              (for what this all is constructed for) or 0.5 in 
                              case of a likelihood fit
    Returns:
        dict
    """
    parnames = inspect.getfullargspec(fn).args
    mindict = dict()
    for i,k in enumerate(parnames):
        mindict[k] = startparams[i]
        if not errors is None: mindict["error_" + k] = errors[i]
        if not limits is None: mindict["limit_" + k] = (limits[i][0], limits[i][1])
    mindict["errordef"] = errordef
    return mindict

class Model(object):
    """
    Describe data with a prediction. The Model class allows to set a function
    for data prediction, and fit it to the data by the means of a chi2 fit.
    It is possible to use a collection of functions to describe a complex model,
    e.g Gaussian + some exponential tail.
    The individual models can be fitted independently, which results in sum_i n_i de
    degrees of freedom for i models with n_i parameters each, or alternatively they c
    can be coupled and share parameters, which results in sum_i n_i - n_ij degrees of
    freedom where n_ij is a shared parameters.
    """

    def __init__(self, func,\
                 startparams=None,\
                 limits=((-np.inf, np.inf),),\
                 errors=(10.,),
                 func_norm=1):
        """
        Args:
            func              (fnc): The function which shall model the data.
                                     It has to be of the form f(x, par1, par2, ...).
                                     Only 1d fits are supported, and "x" must be the
                                     first argument.
        Keyword Args:
            Will be passed to iminuit:    
                startparams (tuple): A set of startparameters. 1 start parameter
                                     per function parameter. A good choice of 
                                     start parameters helps the fit a lot.
                limits      (tuple): individual limit min/max for each parameter
                                     1 tuple (min/max) per parameter
                errors      (tuple): One value per parameter, giving an 1sigma error
                                     estimate
            Additional keywords:
                func_norm   (float): multiply the result of func when evaluated with norm
        """

        # if no startparams are given, construct 
        # and initialize with 0es.
        # FIXME: better estimate?
        if startparams is None:
            startparams = [0]*(func.__code__.co_argcount - 1) # first argument is not a free parameter.

        #def normed_func(*args):
        #    return func_norm*func(*args)

        self._callbacks = [copy_func(func)]
        self.startparams = list(startparams)
        #self.errors = [len(startparams)]
        self.errors = None
        self.n_params = [len(startparams)]
        self.best_fit_params = list(startparams)
        self.coupling_variable = []
        self.all_coupled = False
        self.data = None
        self.data_errs = None
        self.xs = None
        self.chi2_ndf = None
        self.chi2_ndf_components = []
        self.norm = 1
        self.ndf = 1
        self.func_norm = [float(func_norm)]
        self.prediction = lambda xs: reduce(lambda x, y: x + y,\
                                  [self.func_norm[i]*f(xs) for i,f in enumerate(self.components)])
        self.first_guess = None
        self._distribution = None
        self._is_categorical = False 
        self.covariance_matrix = None

    @property
    def distribution(self):
        return self._distribution

    def set_distribution(self, distr):
        """
        Adding a distribution to the model. The distribution shall contain
        the data we want to model.

        Args:
            distr (dashi.histogram):

        Returns:
            None
        """
        self._distribution = distr

    #def construct_minuit_function(self):
    #    func_args, coupling = self.extract_parameters()
    #    parameter_set = ["a","b","c","d","e","f","g"]
    #    if self.all_coupled:
    #        func_args = func_args[0]
    #        # x is always 0
    #        fnc_src = """def to_minimize({}):\n\t
    #        """
    #        #def to_minimize()


    @property
    def n_free_params(self):
        """
        The number of free parameters of this model. The free 
        parameter in a least square fit are 
        number of data points - fit parameters.

        Returns:
            int
        """
        return sum(self.n_free_params)

    def _create_distribution(self, data, bins,\
                             normalize=False, density=True):
        """
        Create a distribution out of the data with the help of 
        dashi.

        Args:
            data        (np.ndarray): input data
            bins (np.ndarray or int): bins to forward to the histogram creation

        Keyword Args:
            normalize         (bool): normalize the resulting distribution by nentries
            density           (bool): if together with normalize, then normalize the 
                                      distribution by area, as for a pdf for example.
                                      The area under the curve will then be unity.
        Returns:
            None
        """
        if np.isscalar(bins):
            bins = np.linspace(min(data), max(data), bins)

        h = d.factory.hist1d(data, bins)

        if normalize:
            h_norm = h.normalized(density=density)
            norm = h.bincontent / h_norm.bincontent
            norm = norm[np.isfinite(norm)][0]
            self.norm = norm
            self.set_distribution(h_norm)
        else:
            self.norm = 1
            self.set_distribution(h)
        self.data = self.distribution.bincontent
        self.xs = self.distribution.bincenters

    def add_first_guess(self, func):
        """
        Use func to estimate better startparameters for the initialization 
        of the fit.

        Args:
            func (callable): The function func has to have the same amount
                             of parameters as we have startparameters.

        Returns:
            None
        """

        assert self.all_coupled or len(self.n_params) == 1, "Does not work yet if not all variables are coupled"
        self.first_guess = func

    def eval_first_guess(self, data):
        """
        Assign a new set of start parameters obtained by calling the
        first geuss metthod

        Args:
            data (np.ndarray): input data, used to evaluate the first guess method.
        Returns:
            None
        """
        assert self.first_guess is not None, "No first guess method provided! Run Model.add_first_guess "
        newstartparams = list(self.first_guess(data))
        assert len(newstartparams) == len(self.startparams), "first guess algorithm must yield startparams!"
        self.startparams = newstartparams

    def couple_models(self, coupling_variable):
        """
        Couple the models by a variable, which means use the variable not
        independently in all model components, but fit it only once.
        E.g. if there are 3 models with parameters p1, p2, k each and they
        are coupled by k, parameters p11, p21, p12, p22, and k will be fitted
        instead of p11, p12, k1, p21, p22, k2.

        Args:
            coupling_variable: variable number of the number in startparams. 
                               This *must* be the index to the respective tuple.

        Returns:
            None
        """
        assert len(np.unique(self.n_params)) == 1,\
            "Models have different numbers of parameters,difficult to couple!"

        self.coupling_variable.append(coupling_variable)

    def construct_error_function(self, startparams, errors, limits, errordef):
        """
        Construct the error function together with the necessary parameters 
        for minuit.
        
        Args:
            startparams (tuple): A set of startparameters. 1 start parameter
                                 per function parameter. A good choice of 
                                 start parameters helps the fit a lot.
            limits      (tuple): individual limit min/max for each parameter
                                 1 tuple (min/max) per parameter
            errors      (tuple): One value per parameter, giving an 1sigma error
                                 estimate
            errordef    (float): The errordef should be 1 for a least square fit 
                                 (for what this all is constructed for) or 0.5 in 
                                 case of a likelihood fit
        Returns:
            tuple (callable, dict)
        """

        concated, concated_pars = concat_functions(self._callbacks)
        error_func = construct_efunc(self.xs, self.data, concated, concated_pars)
        pardict = create_minuit_pardict(error_func, startparams, errors, limits, errordef)
        return error_func, pardict

    def couple_all_models(self):
        """
        "Lock" the model after all components have been added. This 
        will determiine a set of startparameters.
        After this, no other models can be coupled/added any more.

        Returns:
            None
        """
        self.all_coupled = True
        # if self.all_coupled:
        self.startparams = self.startparams[0:self.n_params[0]]

    def __add__(self, other):
        """
        Add another component to the model. This allows for more
        complex functions like gaussian + exponential.

        Args:
            other (Model): A full flexed model which has to be 
                           initialized with its own set of 
                           startparameters

        Returns:
            Model        : A multi-compoment Model
        """
        newmodel = Model(lambda x :x)

        # hack the new model to be like self
        newmodel._callbacks = self._callbacks + other._callbacks
        newmodel.startparams = self.startparams + other.startparams
        newmodel.n_params = self.n_params + other.n_params
        newmodel.best_fit_params = newmodel.startparams
        newmodel.func_norm = self.func_norm + other.func_norm
        return newmodel

    @property
    def components(self):
        lastslice = 0
        thecomponents = []
        for i, cmp in enumerate(self._callbacks):
            thisslice = slice(lastslice, self.n_params[i] + lastslice)
            tmpcmp = copy(cmp) # hack - otherwise it will not work :\
            lastslice += self.n_params[i]
            best_fit = self.best_fit_params[thisslice]
            if self.all_coupled:
                best_fit = self.best_fit_params[0:self.n_params[0]]
            yield lambda xs: self.func_norm[i]*tmpcmp(xs, *best_fit)

    def extract_parameters(self):
        """
        Get the variable names and coupling references for
        the individual model components

        Returns:
            tuple
        """
        all_pars = []
        for i in self._callbacks:
            all_pars.append(iminuit.describe(i))
        return all_pars, self.coupling_variable

    def __call__(self, xs, *params):
        """
        Return the model prediction

        Args:
            xs (np.ndaarray): the values the model should be evaluated on

        Returns:
            np.ndarray
        """
        thecomponents = []
        firstparams = params[0:self.n_params[0]]
        first = self.func_norm[0]*self._callbacks[0](xs, *firstparams)

        lastslice = self.n_params[0]
        for i, cmp in enumerate(self._callbacks[1:]):
            thisslice = slice(lastslice, self.n_params[1:][i] + lastslice)
            # tmpcmp = copy(cmp)
            theparams = list(params[thisslice])
            if self.coupling_variable:
                for k in self.coupling_variable:
                    theparams[k] = firstparams[k]
            elif self.all_coupled:
                theparams = firstparams
            # thecomponents.append(lambda xs: tmpcmp(xs, *params[thisslice]))
            lastslice += self.n_params[1:][i]
            first += self.func_norm[i]*cmp(xs, *theparams)
        return first

    def add_data(self, data,\
                 data_errs=None,\
                 bins=200,\
                 create_distribution=False,\
                 normalize=False,\
                 density=True,\
                 xs=None,\
                 subtract=None):
        """
        Add some data to the model, in preparation for the fit. There are two
        modes of this:
        1) Data needs to be histogrammed, then make sure to set
           'nbins' appropriatly and set the 'create_distribution'
        2) Data needs NOT to be histogrammed. In that case, bins has no meaning
           For a meaningful calculation of chi2, the errors of the data points
           need to be given to data_errs


        Args:
           data (np.array)            : input data

        Keyword Args
           data_errs (np.array)       : errors of the data for chi2 calculation
                                         (only used when not histogramming)
           nbins (int/np.array)       : number of bins or bin array to be passed 
                                         to the histogramming routine
           create_distribution (bool) : data requires the creation of a histogram
                                         first before fitting
           subtract (callable)        : ?
           normalize (bool)           : normalize the data before adding
           density (bool)             : if normalized, assume the data is a pdf.
                                        if False, use bincount for normalization.
        Returns:
           None
        """
        if np.isscalar(bins):
            nbins = bins
        else:
            nbins = len(bins)
        
        if create_distribution:
            self._create_distribution(data, bins, normalize, density=density)
            self._is_categorical = True
            self.ndf = nbins - len(self.startparams)
        else:
            assert xs is not None, "Have to give xs if not histogramming!" 
            if data_errs is not None:
                assert len(data_errs) == len(data), "Data and associated errrors must have the same size"
            else:
                data_errs = np.ones(len(data))
            self.data = data
            self.data_errs = data_errs
            self.xs = xs
            self.ndf = len(data) - len(self.startparams)
            self.bins = None
        if subtract is not None:
            self.data -= subtract(self.xs)

    def fit_to_data(self, silent=False,\
                    use_minuit=True,\
                    errors=None,\
                    limits=None,\
                    errordef=1,\
                    debug_minuit=False,\
                    **kwargs):
        """
        Apply this model to data. This will perform the fit with the help 
        of either minuit or scipy.optimize.

        Args:
            data (np.ndarray)      : the data, unbinned
            silent (bool)          : silence output
            use_minuit (bool)      : use minuit for fitting
            errors (list)          : errors for minuit, see miniuit manual
            limits (list of tuples): limits for minuit, see minuit manual
            errordef (float)       : typically 1 for chi2 fit and 0.5 for llh fit 
                                   : this class is currently set up as a leeast square
                                     fit, so this should not be changed
            debug_minuit (int)     : if True, attache the iminuit instance to the model
                                     so that it can be inspected later on. Will raise error
                                     if use_minuit is set to False at the same time
            **kwargs: will be passed on to scipy.optimize.curvefit

        Returns:
            None
        """
        if not use_minuit and debug_minuit:
            raise ValueError("You can not debug  minuit when you are not using it!")

        startparams = self.startparams

        if not silent: print("Using start params...", startparams)

        fitkwargs = {"maxfev": 1000000, "xtol": 1e-10, "ftol": 1e-10}
        if "bounds" in kwargs:
            fitkwargs.pop("maxfev")
            # this is a matplotlib quirk
            fitkwargs["max_nfev"] = 1000000
        fitkwargs.update(kwargs)
        if use_minuit:
            errorfunc, params = self.construct_error_function(self.startparams,\
                                                               errors,\
                                                               limits,\
                                                               errordef)
            # for debugging reasons we attach the minuit instance
            # to the model if desired
            m = iminuit.Minuit(errorfunc, **params)
            m.migrad()
            values = m.values
            if not silent: print (values, "result")
            parameters=[]
            for k in sorted(m.var2pos, key=m.var2pos.get):
                if not silent : print (k)
                parameters.append(m.values[k])
            self.errors = m.errors
            self.covariance_matrix = m.covariance
        else:
            parameters, covariance_matrix = optimize.curve_fit(self, self.xs,\
                                                           self.data, p0=startparams,\
                                                           # bounds=(np.array([0, 0, 0, 0, 0] + [0]*len(start_params[5:])),\
                                                           # np.array([np.inf, np.inf, np.inf, np.inf, np.inf] +\
                                                           # [np.inf]*len(start_params[5:]))),\
                                                           # max_nfev=100000)
                                                           # method="lm",\
                                                           **fitkwargs)
            self.covariance_matrix = covariance_matrix
            self.errors = [] 
            for i, row in enumerate(self.covariance_matrix):
                for j,entry in enumerate(row):
                    if i == j:
                        self.errors.append(np.sqrt(entry))
        if not silent: print("Fit yielded parameters", parameters)
        if (not silent) and (not use_minuit): print("{:4.2f} NANs in covariance matrix".format(len(self.covariance_matrix[np.isnan(np.asarray(covariance_matrix))])))

        # simple GOF
        #norm = 1
        #if normalize:
        #    norm = h.bincontent / h_norm.bincontent
        #    norm = norm[np.isfinite(norm)][0]

        #self.norm = norm
        if self._is_categorical:
            chi2 = (funcs.calculate_chi_square(self.norm*self.data, self.norm * self(self.xs, *parameters)))
        else:
            # calculate the chi2 
            chi2 = (funcs.calculate_reduced_chi_square(self.norm*self.data, self.norm * self(self.xs, *parameters), self.data_errs))
        self.chi2_ndf = chi2/self.ndf

        # FIXME: new feature
        #for cmp in self.components:
        #    thischi2 = (calculate_chi_square(h.bincontent, norm * cmp(h.bincenters)))
        #    self.chi2_ndf_components.append(thischi2/nbins)
        if not silent and use_minuit: print("Function value at minimum {:4.2e}".format(m.fval))
        if not silent: print("Obtained chi2 : {:4.2f}; ndf : {:4.2f}; chi2/ndf {:4.2f}".format(chi2, self.ndf, self.chi2_ndf))
        if not silent: print("##########################################")
        self.best_fit_params = parameters
        if debug_minuit:
            self._m_instance = m
        return parameters
    
    def get_minuit_instance(self):
        """
        If a previous fit has been done with the debug_minuit instance
        then it now can be accessed.

        """
        if not hasattr(self, '_m_instance'):
            Logger.warn('Minuit instance not available. Execute `fit_to_data` with `debug_minuit` set to True')
            return None
        else:
            return self._m_instance


    def clear(self):
        """
        Reset the model. This bascially deletes all components and 
        resets the startparameters.

        Returns:
            None
        """
        self.__init__(self._callbacks[0], self.startparams[:self.n_params[0]])

    def plot_result(self, ymin=1000, xmax=8,\
                    ylabel="normed bincount",\
                    xlabel="Q [C]", fig=None,\
                    log=True,\
                    figure_factory=None,\
                    axes_range="auto",
                    model_alpha=.3,\
                    add_parameter_text=((r"$\mu_{{SPE}}$& {:4.2e}\\",0),),
                    histostyle="scatter",
                    datacolor="k",
                    modelcolor=default_color):
        """
        Show the fit result, together with the fitted data.

        Args:
            ymin               (float): limit the yrange to ymin
            xmax               (float): limit the xrange to xmax
            model_alpha        (float): 0 <= x <= 1 the alpha value of the lineplot
                                        for the model
            ylabel               (str): label for yaxis
            log                 (bool): plot in log scale
            figure_factory       (fnc): Use to generate the figure
            axes_range           (str): the "field of view" to show
            fig         (pylab.figure): A figure instance
            add_parameter_text (tuple): Display a parameter in the table on the plot
                                        ((text, parameter_number), (text, parameter_number),...)
            datacolor            (str): color for the data points
            modelcolor           (str): color for the model prediction

        Returns:
            pylab.figure
        """
        assert self.chi2_ndf is not None, "Needs to be fitted first before plotting!"

        def auto_adjust_limits(ax, data, xs):
            scalemax, scalemin = 1.1, 0.9

            ymax, ymin = scalemax*max(data), scalemin*min(data)
            xmax, xmin = scalemax*max(xs[abs(data) > 0]), scalemin*min(xs[abs(data) >0])
            ax.set_xlim(xmin, xmax)
            ax.set_ylim(ymin, ymax)
            return ax

        if figure_factory is not None:
            fig = figure_factory()

        elif fig is None:
            fig = p.figure()

        ax = fig.gca()


        if self.distribution is not None:
            self.distribution.__getattribute__(histostyle)(color=datacolor)
        else:
            ax.plot(self.xs, self.data, color=datacolor)

        infotext = r"\begin{tabular}{ll}"

        ax.plot(self.xs, self.prediction(self.xs), color=default_color, alpha=model_alpha)
        for comp in self.components:
            ax.plot(self.xs, comp(self.xs), linestyle=":", lw=1, color="k")

        infotext += r"$\chi^2/ndf$ & {:4.2f}\\".format(self.chi2_ndf)

        ax.set_ylim(ymin=ymin)
        ax.set_xlim(xmax=xmax)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        if axes_range == "auto":
            ax = auto_adjust_limits(ax, self.data, self.xs)

        if self.distribution is not None:
            infotext += r"entries& {}\\".format(self.distribution.stats.nentries)
        if add_parameter_text is not None:
            for partext in add_parameter_text:
                infotext += partext[0].format(self.best_fit_params[partext[1]])
            #infotext += r"$\mu_{{SPE}}$& {:4.2e}\\".format(self.best_fit_params[mu_spe_is_par])
        infotext += r"\end{tabular}"
        ax.text(0.9, 0.9, infotext,
            horizontalalignment='center',
            verticalalignment='center',
            transform=ax.transAxes)
        if log: ax.set_yscale("symlog", linthreshy=1)
        #sb.despine()
        return fig




"""
Provide an easy-to-use, intuitive way of fitting models with different components
to data. The focus is less on a statistical sophisticated fitting rather than on 
an explorative approach to data investigation. This might help answer questions of 
the form - "How compatible is this data with a Gaussian + Exponential?".
Out of the box, this module provides tools targeted to a least-square fit, however,
in principle this could be extended to likelihood fits.

Currently the generation of the minimized error function is automatic, and it is 
generated *only* for the least-squares case, however this might be expanded in the 
future.

"""

from .functions import poisson, gauss, exponential, calculate_chi_square, fwhm_gauss
from .model import Model
